"""
.. _exercise09_label:

Submitting a task to the LSF batch system
=========================================

.. hint::
    This example demonstrates how to submit a task to the LSF batch
    system. In it's core, the task is the same as in :ref:`exercise07_label`.
    The key difference is the ``batch_system`` parameter that is set to ``"lsf"``
    and additional settings that are used to steer the submission to LSF.

To define the batch system on which this task should run, the only
change needed is to set the ``batch_system`` parameter to the desired
batch system. The available options are:

- ``"local"`` for no batch submission

- ``"lsf"`` for LSF (see the `LSF documentation <https://www.ibm.com/docs/en/spectrum-lsf/10.1.0?topic=overview-lsf-introduction>`_)

- ``"htcondor"`` for HTCondor (see the `HTCondor documentation <https://htcondor.readthedocs.io/en/latest/users-manual/quick-start-guide.html>`_)

- ``"slurm"`` for Slurm (see the `Slurm documentation <https://slurm.schedmd.com/quickstart.html>`_)

- ``"gbasf2"`` for the client to submit grid-based Belle II jobs (see the `gbasf2 documentation <https://gbasf2.belle2.org/>`_)

In ``b2luigi`` there are some dedicated parameters for each supported batch system,
see :ref:`batch-label`.
For LSF, these parameters are:

- ``queue``: the queue to submit the job to (default: ``"s"``)

- ``job_name``: the name of the job to be displayed in the queue
"""

import basf2 as b2
import modularAnalysis as ma
import variables.utils as vu
import vertex as vx


import b2luigi
from b2luigi.basf2_helper import Basf2PathTask
from Ex06_basf2_reconstruction import ReconstructionTask


@b2luigi.requires(ReconstructionTask)
class AnalysisTask(Basf2PathTask):
    treeFit = b2luigi.BoolParameter(default=True)
    result_dir = "results/AnalysisLSF"

    batch_system = "lsf"
    queue = "s"

    def output(self):
        yield self.add_to_output(f"ntuple_LSF_{self.identifier}.root")

    def create_path(self):
        main = b2.Path()

        particle = "mu" if str(self.event_type) == "mumu" else "e"

        ma.inputMdstList(
            environmentType="default", filelist=self.get_input_file_names("reconstruction.root"), path=main
        )

        # Fill final state particles
        ma.fillParticleList(decayString=f"{particle}+:fsp", cut="[abs(dr) < 2.0] and [abs(dz) < 4.0]", path=main)

        ma.fillParticleList(
            decayString="K+:fsp", cut="[abs(dr) < 2.0] and [abs(dz) < 4.0] and [kaonID > 0.2]", path=main
        )

        # reconstruct J/psi -> l+ --
        ma.reconstructDecay(
            decayString=f"J/psi:ll -> {particle}+:fsp {particle}-:fsp", cut="[2.92 < M < 3.22]", path=main
        )

        # reconstruct B+ -> J/psi K+
        ma.reconstructDecay(
            decayString="B+:jpsik -> J/psi:ll K+:fsp", cut="[Mbc > 5.24] and [abs(deltaE) < 0.2]", path=main
        )

        # Run truth matching
        ma.matchMCTruth(list_name="B+:jpsik", path=main)

        # Perform tree fit if requested
        if self.treeFit:
            vx.treeFit(list_name="B+:jpsik", conf_level=0.00, massConstraint=["J/psi"], path=main)

        # Define variables to save
        variables = ["Mbc", "deltaE", "chiProb", "isSignal"]
        variables += vu.create_aliases_for_selected(
            list_of_variables=["InvM", "M"], decay_string="B+:jpsik -> ^J/psi:ll K+:fsp", prefix="jpsi"
        )

        # Write variables into ntuple
        ma.variablesToNtuple(
            decayString="B+:jpsik",
            variables=variables,
            filename=self.get_output_file_name(f"ntuple_LSF_{self.identifier}.root"),
            treename="Bp",
            path=main,
            storeEventType=False,
        )

        return main


# %%
# To execute tasks in batch mode, use the ``batch`` argument of the
# ``process`` method.


# %%
if __name__ == "__main__":
    b2luigi.process(AnalysisTask(n_events=10), batch=True)


# %%
# Either pass ``batch=True`` as shown above or use the
# CLI option ``--batch`` when executing the script, e.g.:
#
# .. code-block:: bash
#
#   python3 Ex09_basf2_analysis_LSF.py --batch



import b2luigi

from b2luigi.core.temporary_wrapper import TemporaryFileContextManager

from ..helpers import B2LuigiTestCase


class TemporaryFileContextManagerTestCase(B2LuigiTestCase):
    def test_get_output_file_name(self):
        class TaskA(b2luigi.Task):
            def output(self):
                yield self.add_to_output("final.txt")

            def run(self):
                with open(self.get_output_file_name("final.txt"), "w") as f:
                    f.write("Test")

        task = TaskA()
        non_temp_path = task.get_output_file_name("final.txt")
        with TemporaryFileContextManager(task):
            temp_path = task.get_output_file_name("final.txt")
            self.assertIsInstance(temp_path, str)
            # Using assertNotEqual as a dummy check here. The important part is, that the replaced function is called
            self.assertNotEqual(non_temp_path, temp_path)
            task.run()

    def test_get_input_file_name_simple(self):
        class TaskA(b2luigi.Task):
            def output(self):
                yield self.add_to_output("input.txt")

            @b2luigi.on_temporary_files
            def run(self):
                output_file = self.get_output_file_name("input.txt")
                with open(output_file, "w") as f:
                    f.write("Test")

        task_a = TaskA()
        task_a.run()

        @b2luigi.requires(TaskA)
        class TaskB(b2luigi.Task):
            pass

        task_b = TaskB()

        not_temp_path = task_b.get_input_file_names("input.txt")
        with TemporaryFileContextManager(task_b):
            temp_path = task_b.get_input_file_names("input.txt")
            self.assertIsInstance(temp_path, list)
            # Using assertNotEqual as a dummy check here. The important part is, that the replaced function is called
            self.assertNotEqual(not_temp_path, temp_path)

    def test_get_input_file_names_from_dict(self):
        class TaskA(b2luigi.Task):
            def output(self):
                yield self.add_to_output("input.txt")

            @b2luigi.on_temporary_files
            def run(self):
                output_file = self.get_output_file_name("input.txt")
                with open(output_file, "w") as f:
                    f.write("Test")

        task_a = TaskA()
        task_a.run()

        class TaskB(b2luigi.Task):
            def requires(self):
                return {"req1": TaskA(), "req2": TaskA()}

        task_b = TaskB()
        not_temp_path_req1 = task_b.get_input_file_names_from_dict("req1")
        not_temp_path_req2 = task_b.get_input_file_names_from_dict("req2")
        with TemporaryFileContextManager(task_b):
            temp_path_req1 = task_b.get_input_file_names_from_dict("req1")
            temp_path_req2 = task_b.get_input_file_names_from_dict("req2")
            self.assertIsInstance(temp_path_req1, list)
            self.assertIsInstance(temp_path_req2, list)
            # Using assertNotEqual as a dummy check here. The important part is, that the replaced function is called
            self.assertNotEqual(not_temp_path_req1, temp_path_req1)
            self.assertNotEqual(not_temp_path_req2, temp_path_req2)



import b2luigi


class MyTask(b2luigi.Task):
    def output(self):
        yield self.add_to_output("test.txt")

    @b2luigi.on_temporary_files
    def run(self):
        with open(self.get_output_file_name("test.txt"), "w") as f:
            f.write("Test")


if __name__ == "__main__":
    b2luigi.set_setting("result_dir", "results")
    b2luigi.process(MyTask())



"""
.. _exercise03_label:

``b2luigi.WrapperTask`` and ``b2luigi`` Settings
================================================

.. hint::
    This example demonstrates the usage of the ``b2luigi.WrapperTask`` class.
    This class is used to define a task that requires multiple other tasks to be executed.
    At the end of this section, we introduce the ``b2luigi`` settings mechanism.

The :class:`b2luigi.WrapperTask` class is used to define a task that requires all parameter combinations of another task to be executed.
The combinations are determined by the ``requires`` method of the :class:`b2luigi.WrapperTask`.

To show the functionality of the :class:`b2luigi.WrapperTask` class, we will create a third task that depends on second task and executes all parameter combinations of the second task.
For this, we define the first and second task in the same way as in the previous example.
"""

import b2luigi


class MyTask(b2luigi.Task):
    parameter = b2luigi.IntParameter()

    def run(self):
        with open(self.get_output_file_name("output.txt"), "w") as f:
            f.write(f"{self.parameter}")

    def output(self):
        yield self.add_to_output("output.txt")


class MyOtherTask(b2luigi.Task):
    parameter = b2luigi.IntParameter()

    def requires(self):
        return MyTask(parameter=self.parameter)

    def output(self):
        yield self.add_to_output("output2.txt")

    def run(self):
        with open(self.get_input_file_names("output.txt")[0], "r") as f:
            number = int(f.read())

        with open(self.get_output_file_name("output2.txt"), "w") as f:
            f.write(f"{number**2}")


# %%
# Here, we define the ``b2luigi.WrapperTask`` to loop over all parameter combinations of the second task.
# In contrast to the usual :class:`b2luigi.Task` class, it does not need a ``run`` and ``output`` method.
# The task counts as completed if its requirement is fulfilled.


# %%
class MyWrapperTask(b2luigi.WrapperTask):
    max_value = b2luigi.IntParameter()

    def requires(self):
        return [MyOtherTask(parameter=value) for value in range(self.max_value)]


# %%
# We make use of the settings mechanism of ``b2luigi`` to define the directory where the results will be stored.
# This is done by assigning a value to the `result_dir` key.
# Setting the `results_dir` here, globally, will now save all the outputs of the processed tasks in the specified location.
# It is also possible to define for each task specifically the settings.
# The order in which the settings are respected is as follows:
#
# - Task instance attribute
# - Task class property
# - Global setting (as demonstrated here)
# - Configuration file 'settings.json'
#
# More information about the settings can be found in the :ref:`settings-label` section.

# %%
if __name__ == "__main__":
    b2luigi.set_setting("result_dir", "results")
    b2luigi.process(MyWrapperTask(max_value=10))

# %%
# Check the output with:
#
# .. code-block:: bash
#
#   for file in results/*/output2.txt; do cat "$file"; echo ""; done
#
# In very easy cases, it is also possible to directly process a list of
# tasks at this position. However, this is not recommended for more complex
# task trees. The equivalent process call would be:
#
# .. code-block:: python
#
#   b2luigi.process([MyOtherTask(parameter=value) for value in range(10)])



import json
import re
import subprocess
import os
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_exponential

from b2luigi.batch.processes import BatchProcess, JobStatus
from b2luigi.batch.cache import BatchJobStatusCache
from b2luigi.core.utils import get_log_file_dir
from b2luigi.core.executable import create_executable_wrapper
from b2luigi.core.settings import get_setting


class LSFJobStatusCache(BatchJobStatusCache):
    @retry(
        retry=retry_if_exception_type(subprocess.CalledProcessError),
        stop=stop_after_attempt(4),
        wait=wait_exponential(multiplier=2, min=2, exp_base=3),  # 2, 6, 18 seconds
        reraise=True,
    )
    def _ask_for_job_status(self, job_id=None):
        """
        Queries the job status from the LSF batch system and updates the internal job status mapping.

        Args:
            job_id (str, optional): The ID of the job to query. If not provided,
                                    the status of all jobs will be queried.

        Notes:
            - This method uses the ``bjobs`` command-line tool to fetch job statuses in JSON format.
            - The output is expected to contain a "RECORDS" key with a list of job records.
            - Each job record should have "JOBID" and "STAT" keys, which are used to update the internal mapping.
        """
        if job_id:
            output = subprocess.check_output(["bjobs", "-json", "-o", "jobid stat", str(job_id)])
        else:
            output = subprocess.check_output(["bjobs", "-json", "-o", "jobid stat"])
        output = output.decode()
        output = json.loads(output)["RECORDS"]

        for record in output:
            self[record["JOBID"]] = record["STAT"]


_batch_job_status_cache = LSFJobStatusCache()


class LSFProcess(BatchProcess):
    """
    Reference implementation of the batch process for a LSF batch system.
    Heavily inspired by `this post <https://github.com/spotify/luigi/pull/2373/files>`_.

    Additional to the basic batch setup (see :ref:`batch-label`), there are
    LSF-specific :meth:`settings <b2luigi.set_setting>`. These are:

    * the LSF queue: ``queue``.
    * the number of slots for the job. On KEKCC this increases the memory available to the job: ``job_slots``.
    * the LSF job name: ``job_name``.

    For example:

    .. code-block:: python

        class MyLongTask(b2luigi.Task):
            queue = "l"
            job_name = "my_long_task"

    The default queue is the short queue ``"s"``. If no ``job_name`` is set the task
    will appear as ::

        <result_dir>/parameter1=value/.../executable_wrapper.sh"

    when running ``bjobs``.

    By default, the environment variables from the scheduler are copied to
    the workers.
    This also implies we start in the same working directory, can reuse
    the same executable, etc.
    Normally, you do not need to supply ``env_script`` or alike.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._batch_job_id = None

    def get_job_status(self):
        """
        Retrieves the current status of the batch job associated with this instance.

        Returns:
            JobStatus: The status of the job, which can be one of the following:
                - :meth:`JobStatus.successful <b2luigi.process.JobStatus.successful>`: If the job has completed successfully ("DONE").
                - :meth:`JobStatus.aborted <b2luigi.process.JobStatus.aborted>`: If the job has been aborted or is not found in the cache ("EXIT" or missing ID).
                - :meth:`JobStatus.running <b2luigi.process.JobStatus.running>`: If the job is still in progress.
        """
        if not self._batch_job_id:
            return JobStatus.aborted

        try:
            job_status = _batch_job_status_cache[self._batch_job_id]
        except KeyError:
            return JobStatus.aborted

        if job_status == "DONE":
            return JobStatus.successful
        if job_status == "EXIT":
            return JobStatus.aborted

        return JobStatus.running

    def start_job(self):
        """
        Submits a batch job to the LSF system.

        This method constructs a command to submit a job using the ``bsub`` command-line tool.
        It dynamically configures the job submission parameters based on task-specific settings
        and creates necessary log files for capturing standard output and error.

        Raises:
            RuntimeError: If the batch submission fails or the job ID cannot be extracted
                          from the ``bsub`` command output.

        Steps:
            1. Retrieve optional settings for ``queue`` (``-q``), ``job_slots`` (``-n``), and ``job_name`` (``-J``).
            2. The ``stdout`` and ``stderr`` log files are created in the task's log directory. See :obj:`get_log_file_dir`.
            3. The executable is created with :obj:`create_executable_wrapper`.
        """
        command = ["bsub", "-env all"]

        queue = get_setting("queue", task=self.task, default=False)
        if queue is not False:
            command += ["-q", queue]

        job_slots = get_setting("job_slots", task=self.task, default=False)
        if job_slots is not False:
            command += ["-n", job_slots]

        job_name = get_setting("job_name", task=self.task, default=False)
        if job_name is not False:
            command += ["-J", job_name]

        log_file_dir = get_log_file_dir(self.task)
        os.makedirs(log_file_dir, exist_ok=True)

        stdout_log_file = os.path.join(log_file_dir, "stdout")
        stderr_log_file = os.path.join(log_file_dir, "stderr")

        command += ["-eo", stderr_log_file, "-oo", stdout_log_file]

        executable_file = create_executable_wrapper(self.task)
        command.append(executable_file)

        output = subprocess.check_output(command)
        output = output.decode()

        # Output of the form Job <72065926> is submitted to default queue <s>.
        match = re.search(r"<[0-9]+>", output)
        if not match:
            raise RuntimeError("Batch submission failed with output " + output)

        self._batch_job_id = match.group(0)[1:-1]

    def terminate_job(self):
        """
        This method checks if a batch job ID is set. If it exists, it attempts to
        terminate the job using the ``bkill`` command. The command's output is suppressed,
        and errors during execution are not raised.
        """
        if not self._batch_job_id:
            return

        subprocess.run(["bkill", self._batch_job_id], stdout=subprocess.DEVNULL, check=False)



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import pathlib
import sys

sys.path.insert(0, pathlib.Path(__file__).parents[1].resolve().as_posix())


# -- Project information -----------------------------------------------------


def get_latest_tag():
    """
    Helper function that returns the name of the tag if the latest commit is tagged,
    otherwise it simply returns "latest"
    """

    default = "latest"

    try:
        import git
    except ImportError:
        print("GitPython module is not installed")
        return default

    current_directory = os.getcwd()

    while True:
        try:
            repo = git.Repo(current_directory)
            latest_commit = repo.head.commit
            tags = repo.tags
            for tag in tags:
                if tag.commit == latest_commit:
                    return tag.name
            # If the latest commit has not tags, then it's the latest version
            return default
        except git.InvalidGitRepositoryError:
            # No Git repository found in the current directory
            parent_directory = os.path.dirname(current_directory)
            # Check if we have reached the root directory: if yes, break the loop and then return
            if parent_directory == current_directory:
                break
            # Move to the parent directory
            current_directory = parent_directory

    print("GitPython did not find a valid repository")
    return default


project = "b2luigi"
copyright = "2018-2023, Nils Braun, Michael Eliachevitch, The Belle II Collaboration"
author = "Nils Braun, Michael Eliachevitch, The Belle II Collaboration"

# The short X.Y version
_version = get_latest_tag()
version = _version
# The full version, including alpha/beta/rc tags
release = _version


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_gallery.gen_gallery",
    "sphinx.ext.viewcode",
    "sphinx_sitemap",
    "notfound.extension",
]

sphinx_gallery_conf = {
    "examples_dirs": "../examples/tutorial",
    "gallery_dirs": "starterkit",
    "filename_pattern": r"Ex\d+.*\.py",
    "ignore_pattern": r"^.*\.txt$|.*\.sh$",
    "plot_gallery": False,
    "show_memory": False,
    "min_reported_time": 0,
    "remove_config_comments": True,
    "backreferences_dir": None,
    "download_all_examples": False,
    "notebook_extensions": {},  # Don't create notebook download
    "within_subsection_order": "FileNameSortKey",
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = {".rst": "restructuredtext"}

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

autodoc_member_order = "bysource"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]
html_logo = "b2luigi.png"

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# Notfound extension options
notfound_context = {"title": "Redirecting...", "body": '<meta http-equiv="refresh" content="0; url=/" />'}

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
html_baseurl = "https://b2luigi.belle2.org/"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "b2luigi.tex",
        "b2luigi Documentation",
        "Nils Braun, Michael Eliachevitch, The Belle II Collaboration",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "b2luigi", "b2luigi Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "b2luigi",
        "b2luigi Documentation",
        author,
        "b2luigi",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# Cheating around read the docs: we need to install our project which is not possible because we
# have no setup.py. But this should work...

if os.getenv("READTHEDOCS"):
    import subprocess

    subprocess.check_call(["pip", "install", "flit"])
    subprocess.check_call(["flit", "install", "-s"], cwd="../")



import json
import os
import warnings
from unittest.mock import patch
from ..helpers import B2LuigiTestCase

import b2luigi
from b2luigi.core.settings import DeprecatedSettingsWarning


class TaskTestCase(B2LuigiTestCase):
    def setUp(self):
        super().setUp()

        b2luigi.clear_setting("my_setting")

    def test_set_by_function(self):
        self.assertRaises(ValueError, b2luigi.get_setting, "my_setting")

        self.assertEqual("default", b2luigi.get_setting("my_setting", "default"))

        b2luigi.set_setting("my_setting", "my value")

        self.assertEqual("my value", b2luigi.get_setting("my_setting"))
        self.assertEqual("my value", b2luigi.get_setting("my_setting", "default"))

    def test_set_by_file(self):
        with open("settings.json", "w") as f:
            json.dump({"my_setting": "my file value"}, f)

        self.assertEqual("my file value", b2luigi.get_setting("my_setting"))

        b2luigi.set_setting("my_setting", "my value")

        self.assertEqual("my value", b2luigi.get_setting("my_setting"))

    def test_set_by_parent_file(self):
        os.mkdir("child_path")
        os.chdir("child_path")

        with open("../settings.json", "w") as f:
            json.dump({"my_setting": "my parent value"}, f)

        self.assertEqual("my parent value", b2luigi.get_setting("my_setting"))

        with open("settings.json", "w") as f:
            json.dump({"my_setting": "my child value"}, f)

        self.assertEqual("my child value", b2luigi.get_setting("my_setting"))

    def test_set_by_task_or_file(self):
        with open("settings.json", "w") as f:
            json.dump({"my_setting": "my file value"}, f)

        b2luigi.set_setting("my_second_setting", "my value")

        task = b2luigi.Task()
        setattr(task, "my_third_setting", "my task value")

        self.assertEqual("my file value", b2luigi.get_setting("my_setting", task=task))
        self.assertEqual("my value", b2luigi.get_setting("my_second_setting", task=task))
        self.assertEqual("my task value", b2luigi.get_setting("my_third_setting", task=task))

    def test_deprecated_settings(self):
        self.assertRaises(ValueError, b2luigi.get_setting, key="my_setting", deprecated_keys=["my_old_setting"])

        b2luigi.set_setting("my_old_setting", "my value")

        with warnings.catch_warnings(record=True) as w:
            self.assertEqual("my value", b2luigi.get_setting("my_setting", deprecated_keys=["my_old_setting"]))

            self.assertEqual(len(w), 1)
            self.assertIsInstance(w[-1].message, DeprecatedSettingsWarning)
            self.assertIn("deprecated", str(w[-1].message))

        b2luigi.set_setting("my_setting", "my new_value")

        with warnings.catch_warnings(record=True) as w:
            self.assertEqual(
                "my new_value", b2luigi.get_setting("my_setting", default="default", deprecated_keys=["my_old_setting"])
            )

            self.assertEqual(len(w), 0)

    def test_b2luigi_settings_json(self):
        with patch.dict("os.environ", {"B2LUIGI_SETTINGS_JSON": "settings.json"}):
            self.assertEqual("settings.json", next(b2luigi._setting_file_iterator()))

    def test_task_property_with_attribute_error(self):
        """Test that AttributeError in task property is raised.

        This test covers the case where a task property causes an AttributeError
        (e.g., accessing a non-existent attribute). The get_setting function should
        raise the AttributeError instead of catching it.
        """

        class MyTask(b2luigi.Task):
            param = b2luigi.Parameter()

            @property
            def foo(self):
                # This will raise AttributeError when accessed
                return self.xparam

        task = MyTask(param="test_value")

        self.assertRaises(AttributeError, b2luigi.get_setting, "foo", task=task)



from typing import Iterable, Iterator
from b2luigi.core import utils
from typing import Any, Union, List, Dict, Optional, Type

import luigi

from b2luigi.core.utils import create_output_file_name
from b2luigi.core.target import LocalTarget, FileSystemTarget
from b2luigi.core.settings import get_setting, _no_value


class Task(luigi.Task):
    """
    Drop in replacement for ``luigi.Task`` which is 100% API compatible.
    It just adds some useful methods for handling output file name generation using
    the parameters of the task.
    See :ref:`quick-start-label` on information on how to use the methods.

    Example:

        .. code-block:: python

          class MyAverageTask(b2luigi.Task):
              def requires(self):
                  for i in range(100):
                      yield self.clone(MyNumberTask, some_parameter=i)

              def output(self):
                  yield self.add_to_output("average.txt")

              def run(self):
                  # Build the mean
                  summed_numbers = 0
                  counter = 0
                  for input_file in self.get_input_file_names("output_file.txt"):
                      with open(input_file, "r") as f:
                          summed_numbers += float(f.read())
                          counter += 1

                  average = summed_numbers / counter

                  with self.get_output_file("average.txt").open("w") as f:
                      f.write(f"{average}\\n")
    """

    def add_to_output(
        self,
        output_file_name: str,
        target_class: Optional[Type[FileSystemTarget]] = None,
        result_dir: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, LocalTarget]:
        """
        Call this in your ``output()`` function to add a target to the list of files,
        this task will output.
        Always use in combination with ``yield``.
        This function will automatically add all current parameter values to
        the file name when used in the form::

            <result-path>/param1=value1/param2=value2/.../<output-file-name.ext>

        This function will by default use a :class:`b2luigi.LocalTarget`, but you can also pass a different ``target_class`` as an argument.
        If you do not want this, you can override the :obj:`_get_output_file_target` function.

        Example:
            This adds two files called ``some_file.txt`` and ``some_other_file.txt`` to the output:

            .. code-block:: python

              def output(self):
                  yield self.add_to_output("some_file.txt")
                  yield self.add_to_output("some_other_file.txt")

        Args:
            output_file_name (str): the file name of the output file.
                Refer to this file name as a key when using :obj:`get_input_file_names`,
                :obj:`get_output_file_names` or :obj:`get_output_file`.
            target_class: which class of :obj:`FileSystemTarget` to instantiate for this target.
                defaults to :class:`b2luigi.LocalTarget`
            result_dir (str, optional): Optionally pass a `result_dir` to the :obj:`create_output_file_name`.
            kwargs: kwargs to be passed to the `__init__` of the Target_class via the :obj:`_get_output_file_target` function

        Returns:
            A dictionary with the output file name as key and the target as value.
        """
        return {
            output_file_name: self._get_output_file_target(
                output_file_name, target_class, result_dir=result_dir, **kwargs
            )
        }

    @staticmethod
    def _transform_io(input_generator: Iterable[FileSystemTarget]) -> Dict[str, List[str]]:
        file_paths: Dict[str, List[str]] = utils.flatten_to_file_paths(input_generator)

        return file_paths

    def get_all_input_file_names(self) -> Iterator[str]:
        """
        Return all file paths required by this task.

        Example:

            .. code-block:: python

              class TheSuperFancyTask(b2luigi.Task):
                  def dry_run(self):
                      for name in self.get_all_output_file_names():
                          print(f"\t\toutput:\t{name}")

        Yields:
            Iterator[str]: An iterator over the input file paths as strings.
        """
        for file_names in self._transform_io(self.input()).values():
            for file_name in file_names:
                yield file_name

    def get_input_file_names(self, key: Optional[str] = None) -> Union[Dict[str, List[str]], List[str]]:
        """
        Get a dictionary of input file names of the tasks, which are defined in our requirements.
        Either use the key argument or dictionary indexing with the key given to :obj:`add_to_output`
        to get back a list (!) of file paths.

        Args:
            key (str, optional): If given, only return a list of file paths with this given key.

        Return:
            If key is none, returns a dictionary of keys to list of file paths.
            Else, returns only the list of file paths for this given key.
        """
        if key is not None:
            return self._transform_io(self.input())[key]
        return self._transform_io(self.input())

    def get_input_file_names_from_dict(
        self, requirement_key: str, key: Optional[str] = None
    ) -> Union[Dict[str, List[str]], List[str]]:
        """
        Get a dictionary of input file names of the tasks, which are defined in our requirements.

        The requirement method should return a dict whose values are generator expressions (!)
        yielding required task objects.

        Example:

            .. code-block:: python

              class TaskB(luigi.Task):

                  def requires(self):
                      return {
                          "a": (TaskA(5.0, i) for i in range(100)),
                          "b": (TaskA(1.0, i) for i in range(100)),
                      }

                  def run(self):
                      result_a = do_something_with_a(
                          self.get_input_file_names_from_dict("a")
                      )
                      result_b = do_something_with_b(
                          self.get_input_file_names_from_dict("b")
                      )

                      combine_a_and_b(
                          result_a,
                          result_b,
                          self.get_output_file_name("combined_results.txt")
                      )

                  def output(self):
                      yield self.add_to_output("combined_results.txt")

        Either use the key argument or dictionary indexing with the key given to :obj:`add_to_output`
        to get back a list (!) of file paths.

        Args:
            requirement_key (str): Specifies the required task expression.
            key (str, optional): If given, only return a list of file paths with this given key.

        Return:
            If key is none, returns a dictionary of keys to list of file paths.
            Else, returns only the list of file paths for this given key.
        """
        if key is not None:
            return self._transform_io(self.input()[requirement_key])[key]
        return self._transform_io(self.input()[requirement_key])

    def get_input_file_name(self, key: Optional[str] = None) -> str:
        """
        Wraps :obj:`get_input_file_names` and asserts there is only one input file.

        Args:
            key (str, optional): Return the file path with this given key.

        Return:
            File path for the given key.
        """
        input_obj = self.get_input_file_names(key)
        if isinstance(input_obj, list):
            if len(input_obj) == 1:
                return input_obj[0]
        elif isinstance(input_obj, dict):
            if len(input_obj) == 1:
                value = next(iter(input_obj.values()))
                if isinstance(value, list) and len(value) == 1:
                    return value[0]
        raise ValueError(
            f"Found more than 1 input file for the key '{key}'. If this is expected use self.get_input_file_names instead."
        )

    def get_all_output_file_names(self) -> Iterator[str]:
        """
        Return all file paths created by this task.

        Example:

            .. code-block:: python

              class TheSuperFancyTask(b2luigi.Task):
                  def dry_run(self):
                      for name in self.get_all_output_file_names():
                          print(f"\t\toutput:\t{name}")

        Yields:
            str: The file path of each output file.
        """
        for file_names in self._transform_io(self.output()).values():
            for file_name in file_names:
                yield file_name

    def get_output_file_name(self, key: str) -> str:
        """
        Analogous to :obj:`get_input_file_names` this function returns
        a an output file defined in out output function with
        the given key.

        In contrast to :obj:`get_input_file_names`, only a single file name
        will be returned (as there can only be a single output file with a given name).

        Args:
            key (str): Return the file path with this given key.

        Return:
            Returns only the file path for this given key.
        """
        target: luigi.Target = self._get_output_target(key)
        file_path: str = target.path

        return file_path

    def _get_input_targets(self, key: str) -> List[luigi.Target]:
        """
        Retrieve the input targets associated with a specific key.

        This method acts as a shortcut to access the input targets for a given key
        from the task's input.

        Args:
            key (str): The key for which to retrieve the corresponding input targets.

        Returns:
            luigi.Target: The luigi target(s) associated with the specified key.

        Raises:
            KeyError: If the specified key is not found in the input dictionary.
        """
        input_dict = utils.flatten_to_dict_of_lists(self.input())
        return input_dict[key]

    def _get_output_target(self, key: str) -> FileSystemTarget:
        """
        Retrieves the output target associated with a specific key.

        This method acts as a shortcut to access a ``luigi`` target from the task's output.

        Args:
            key (str): The key for which the output target is to be retrieved.

        Returns:
            luigi.Target: The ``luigi`` target associated with the specified key.
        """
        output_dict: Dict[str, FileSystemTarget] = utils.flatten_to_dict(self.output())
        return output_dict[key]

    def _get_output_file_target(
        self,
        base_filename: str,
        target_class: Optional[Type[FileSystemTarget]] = None,
        result_dir: Optional[str] = None,
        **kwargs: Any,
    ) -> FileSystemTarget:
        """
        Generates a Luigi file system target for the output file.

        This method constructs the output file name using the provided base filename
        and additional keyword arguments, and then returns a file system target
        instance of the specified target class.

        Args:
            base_filename (str): The base name of the output file.
            target_class (Type[FileSystemTarget], optional): The class of the file system target to use.
                Defaults to :class:`b2luigi.LocalTarget`.
            result_dir (str, optional): Optionally pass a `result_dir` to the :obj:`create_output_file_name`.
            kwargs (Any): Additional keyword arguments passed to the target_class' `__init__`

        Returns:
            LocalTarget: An instance of the specified file system target class pointing to the output file.
        """
        file_name: str = create_output_file_name(
            self,
            base_filename,
            result_dir=result_dir,
        )
        target = self._resolve_output_file_target(
            file_name,
            target_class=target_class,
            **kwargs,
        )
        return target

    def _resolve_output_file_target(
        self,
        filename: str,
        target_class: Optional[Type[FileSystemTarget]] = None,
        **kwargs: Any,
    ) -> FileSystemTarget:
        """
        Resolves and returns a file system target based on the provided filename and optional target class.
        This method determines the appropriate file system target class to use for the given filename.
        It prioritizes the provided `target_class`, falls back to a default target class specified in
        the settings, or defaults to using a :class:`b2luigi.LocalTarget` if no other target class is specified.

        Args:
            filename (str): The name of the file for which the target is being resolved.
            target_class (Optional[Type[FileSystemTarget]]): An optional file system target class to use.
              If provided, this class will be used to create the target.
            kwargs (Any): Additional keyword arguments to pass to the target class constructor.

        Returns:
            FileSystemTarget: The resolved file system target class.

        Notes:
            - The `default_task_target_class` and `target_class_kwargs` settings are retrieved using
              the :meth:`b2luigi.get_setting` function.
            - If `target_class` is provided, it takes precedence over the default target class.
            - If neither `target_class` nor `default_task_target_class` is provided, a :class:`b2luigi.LocalTarget`
              is used as the default.
        """

        setting_target_class = get_setting("default_task_target_class", default=_no_value, task=self)
        setting_target_class_kwargs = get_setting("target_class_kwargs", default=_no_value, task=self)

        if setting_target_class_kwargs is not _no_value:
            kwargs.update(setting_target_class_kwargs)

        # Target class was already given
        if target_class is not None:
            return target_class(filename, **kwargs)

        # Neither target class nor default_task_target_class was given
        # Use the local target
        if setting_target_class is _no_value:
            return LocalTarget(filename, **kwargs)

        # Finally, use the given target class
        return setting_target_class(filename, **kwargs)

    def _remove_output_file_target(self, base_filename: str) -> None:
        """
        Removes the output file target associated with the given base filename.

        This method retrieves the output file target using the provided base filename
        and attempts to remove it. If the target does not have a `remove` method,
        a `NotImplementedError` is raised.

        Args:
            base_filename (str): The base filename used to identify the output file target.

        Raises:
            NotImplementedError: If the target does not have a `remove` method.
        """
        target: FileSystemTarget = self._get_output_file_target(base_filename)
        if hasattr(target, "remove"):
            target.remove()
        else:
            raise NotImplementedError(
                f"Cannot remove output file target for {base_filename}. " "The target does not have a remove method."
            )

    def _remove_output(self) -> None:
        """
        Removes all output file targets associated with the task.

        This method iterates through all output file names retrieved from
        :obj:`get_all_output_file_names` and removes each corresponding output
        file target by calling :obj:`_remove_output_file_target`.

        .. warning::
            Be very careful with this method!
            It will remove all output files of this task!
            This is not reversible.

        .. hint::
            If you are very sure in what you are doing, you can use this method
            to remove all output files of this task by calling it in the
            :obj:`remove_output` method of your task.

        Example:
            .. code-block:: python

              class TheSuperFancyTask(b2luigi.Task):
                  def remove_output(self):
                      self._remove_output()
        Returns:
            None
        """
        for key in self.get_all_output_file_names():
            try:
                self._remove_output_file_target(key)
            except Exception as ex:
                print(f"Could not remove output file {key}: {ex}")


class ExternalTask(Task, luigi.ExternalTask):
    """Direct copy of :obj:`luigi.ExternalTask`, but with the capabilities of :obj:`Task` added."""

    pass


class WrapperTask(Task, luigi.WrapperTask):
    """Direct copy of :obj:`luigi.WrapperTask`, but with the capabilities of :obj:`Task` added."""

    pass


class NotCompletedTask(Task):
    """
    A custom task class that extends the base Task class and provides a
    specialized `complete` method to determine task completion status.
    This class introduces an additional `check_complete` attribute to
    control whether child tasks are checked for completion.

    Attributes:
        check_complete (bool): A flag indicating whether to check the
            completion status of child tasks. Defaults to True.

    Methods:
        complete() -> bool:
            Determines if the task is complete. This method checks the
            completion status of the current task and its child tasks
            (if `check_complete` is True). If any child task is incomplete,
            the method returns False. If `check_complete` is False, the
            method assumes the task is complete regardless of child task
            statuses.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self.check_complete: bool = True

    def complete(self) -> bool:
        """
        Determines whether the task and its dependencies are complete.

        This method overrides the default ``complete`` method to include a custom
        check for child tasks. It recursively checks the completion status of
        required tasks until a task with ``check_complete = False`` is encountered.

        Returns:
            bool: True if the task and all its dependencies are complete,
                  False otherwise.

        Raises:
            AttributeError: If the ``requires`` method does not return a single task
                            or an iterable of tasks.
        """
        if not super().complete():
            return False

        if not self.check_complete:
            return True

        requires = self.requires()

        try:
            if not requires.complete():
                return False
        except AttributeError:
            for task in requires:
                if not task.complete():
                    return False

        return True



import b2luigi


class MyTask(b2luigi.Task):
    """
    A simple task that does nothing but can be used to test dry run.
    """

    def output(self):
        return b2luigi.LocalTarget("my_task_output.txt")

    def dry_run(self):
        print("This is a dummy output.\n")


if __name__ == "__main__":
    b2luigi.process(MyTask())



import collections
import colorama
import copy
import itertools
import logging
import os
import shlex
import shutil
import sys
import types
from typing import Any, Dict, List, Optional, Iterator, Iterable

import luigi

from b2luigi.core.settings import get_setting


def product_dict(**kwargs: Any) -> Iterator[Dict[str, Any]]:
    """
    Cross-product the given parameters and return a list of dictionaries.

    Example:

        .. code-block:: python

            >>> list(product_dict(arg_1=[1, 2], arg_2=[3, 4]))
            [{'arg_1': 1, 'arg_2': 3}, {'arg_1': 1, 'arg_2': 4}, {'arg_1': 2, 'arg_2': 3}, {'arg_1': 2, 'arg_2': 4}]

        The thus produced list can directly be used as inputs for a required tasks:

        .. code-block:: python

            def requires(self):
                for args in product_dict(arg_1=[1, 2], arg_2=[3, 4]):
                    yield some_task(**args)

    Args:
        kwargs (Any): Each keyword argument should be an iterable

    Return:
        Iterator[Dict[str, Any]]: A list of kwargs where each list of input keyword
        arguments is cross-multiplied with every other.
    """
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in itertools.product(*vals):
        yield dict(zip(keys, instance))


def fill_kwargs_with_lists(**kwargs: Any) -> Dict[str, List[Any]]:
    """
    Return the kwargs with each value mapped to [value] if not a list already.

    Example:
    .. code-block:: python

        >>> fill_kwargs_with_lists(arg_1=[1, 2], arg_2=3)
        {'arg_1': [1, 2], 'arg_2': [3]}

    Args:
        kwargs: The input keyword arguments
    Return:
        Dict[str, List[Any]]: Same as kwargs, but each value mapped to a list if not a list already
    """
    return_kwargs = {}
    for key, value in kwargs.items():
        if value is None:
            value = []
        if isinstance(value, str) or not isinstance(value, collections.abc.Iterable):
            value = [value]
        return_kwargs[key] = value

    return return_kwargs


def flatten_to_file_paths(inputs: Iterable[luigi.target.FileSystemTarget]) -> Dict[str, List[str]]:
    """
    Take in a dict of lists of luigi targets and replace each luigi target by its corresponding path.
    For dicts, it will replace the value as well as the key. The key will however only by the basename of the path.

    Args:
        inputs: A dict of lists of luigi targets
    Return:
        Dict[str, List[str]]: A dict with the keys replaced by the basename of the targets and the values by the full path
    """
    input_dict: Dict[Any, List[luigi.target.FileSystemTarget]] = flatten_to_dict_of_lists(inputs)

    return {
        os.path.basename(key.path if hasattr(key, "path") else key): [
            val.path if hasattr(val, "path") else val for val in value
        ]
        for key, value in input_dict.items()
    }


def flatten_to_dict(inputs: Iterable[Any]) -> Dict[Any, Any]:
    """
    Return a whatever input structure into a dictionary.
    If it is a dict already, return this.
    If is is an iterable of dict or dict-like objects, return the merged dictionary.
    All non-dict values will be turned into a dictionary with value -> ``{value: value}``

    Example:
    .. code-block:: python

        >>> flatten_to_dict([{"a": 1, "b": 2}, {"c": 3}, "d"])
        {'a': 1, 'b': 2, 'c': 3, 'd': 'd'}

    Args:
        inputs (Iterable[Any]): The input structure
    Return:
        Dict[Any, Any]: A dict constructed as described above.
    """
    inputs: List[Any] = _flatten(inputs)
    inputs: Dict[Any, Any] = map(_to_dict, inputs)

    joined_dict = {}
    for i in inputs:
        for key, value in i.items():
            joined_dict[key] = value
    return joined_dict


def flatten_to_dict_of_lists(inputs: Iterable[Any]) -> Dict[Any, List]:
    """
    Flattens a nested iterable structure into a dictionary of lists.

    This function takes an iterable of potentially nested structures, flattens it,
    and converts it into a dictionary where each key maps to a list of values
    associated with that key.

    Args:
        inputs (Iterable[Any]): The input iterable containing nested structures.

    Returns:
        Dict[Any, List]: A dictionary where keys are derived from the input
        and values are lists of corresponding items.
    """
    inputs: List = _flatten(inputs)
    inputs: Dict[List] = map(_to_dict, inputs)

    joined_dict: Dict[str, List] = collections.defaultdict(list)
    for i in inputs:
        for key, value in i.items():
            joined_dict[key].append(value)
    return dict(joined_dict)


def task_iterator(task, only_non_complete=False):
    """
    Iterates through a task and its dependencies in a directed acyclic graph (DAG),
    ensuring that each task is yielded only once.

    Args:
        task: The root task to start iterating from. This task should have methods
              ``complete()`` to check if the task is complete and ``deps()`` to retrieve
              its dependencies.
        only_non_complete (bool, optional): If True, only tasks that are not complete
              (as determined by the ``complete()`` method) will be yielded. Defaults to ``False``.

    Yields:
        task: Each unique task in the DAG, starting from the given root task and
              including its dependencies.

    Notes:
        - The function uses a cache (``already_seen_tasks``) to ensure that tasks
          are not yielded multiple times, even if they are dependencies of multiple
          parent tasks in the DAG.
        - The iteration is performed recursively using a nested helper function.
    """
    already_seen_tasks = set()

    # create another private function for recursion that has reference to `already_seen_tasks`
    def _unique_task_iterator(task, only_non_complete):
        if only_non_complete and task.complete():
            return

        yield task

        for dep in task.deps():
            if dep.task_id in already_seen_tasks:
                continue
            already_seen_tasks.add(dep.task_id)
            yield from _unique_task_iterator(dep, only_non_complete=only_non_complete)

    yield from _unique_task_iterator(task, only_non_complete)


def get_all_output_files_in_tree(root_module, key=None):
    """
    Recursively retrieves all output files from tasks in a given module tree.

    This function iterates through all tasks in the specified root module,
    collects their output files, and organizes them into a dictionary. If a
    specific key is provided, it returns the output files corresponding to
    that key.

    Args:
        root_module (module): The root module containing tasks to iterate over.
        key (str, optional): A specific key to filter the output files. If
            provided, only the output files corresponding to this key will
            be returned.

    Returns:
        dict: A dictionary where keys are file identifiers and values are lists
        of dictionaries containing:

            - ``exists`` (bool): Whether the file exists.
            - ``parameters`` (dict): Serialized parameters of the task.
            - ``file_name`` (str): Absolute path to the file.

    Raises:
        KeyError: If the specified key is not found in the output files.

    Notes:
        - The function uses :obj:`task_iterator` to traverse tasks in the module tree.
        - Output files are flattened and converted to absolute file paths.
    """
    if key:
        return get_all_output_files_in_tree(root_module)[key]

    all_output_files = collections.defaultdict(list)
    for task in task_iterator(root_module):
        output_dict = flatten_to_dict(task.output())
        if not output_dict:
            continue

        for target_key, target in output_dict.items():
            converted_dict = flatten_to_file_paths({target_key: target})
            file_key, file_name = converted_dict.popitem()

            all_output_files[file_key].append(
                dict(
                    exists=target.exists(),
                    parameters=get_serialized_parameters(task),
                    file_name=os.path.abspath(file_name.pop()),
                )
            )

    return all_output_files


def get_serialized_parameters(task, need_hidden=False):
    """
    Retrieve a string-typed ordered dictionary of significant parameters in the format ``key=value``.

    This function iterates over the parameters of a given task, filters out non-significant parameters,
    and serializes the significant ones. If a parameter has a custom serialization method (``serialize_hashed``),
    it will use that; otherwise, it defaults to the standard ``serialize`` method.

    Args:
        task: An object representing a task, which must implement the ``get_params`` method to
              retrieve its parameters and their metadata.

    Returns:
        collections.OrderedDict: An ordered dictionary where keys are parameter names and values
                                 are their serialized representations.
    """
    serialized_parameters = collections.OrderedDict()

    for key, parameter in task.get_params():
        if not parameter.significant or (parameter.hidden and not need_hidden):
            continue

        value = getattr(task, key)
        if hasattr(parameter, "serialize_hashed"):
            value = parameter.serialize_hashed(value)
        else:
            value = parameter.serialize(value)

        serialized_parameters[key] = value

    return serialized_parameters


def create_output_file_name(
    task, base_filename: str, result_dir: Optional[str] = None, need_hidden: bool = False
) -> str:
    """
    Generates an output file path based on the task's parameters, a base filename,
    and an optional result directory.

    Args:
        task: The task object containing parameters to serialize and use in the output path.
        base_filename (str): The base name of the output file.
        result_dir (Optional[str]): The directory where the output file should be saved.
            If not provided, it defaults to the ``result_dir`` setting.

    Returns:
        str: The full path to the output file.

    Raises:
        ValueError: If any parameter value contains a path separator or cannot be
            interpreted as a valid directory name.

    Notes:
        - The function ensures that the result directory is evaluated relative to the
          current executed file.
        - The parameter separator and whether to include parameter names in the output
          path are configurable via settings.
        - If ``use_parameter_name_in_output`` is enabled, the output path includes
          parameter names and values; otherwise, only parameter values are used.
        - If ``parameter_separator`` is set to a non-empty string, it will be used to separate
          parameter names and values in the output path.
    """
    serialized_parameters = get_serialized_parameters(task, need_hidden=need_hidden)

    if not result_dir:
        # Be sure to evaluate things relative to the current executed file, not to where we are now
        result_dir: str = map_folder(get_setting("result_dir", task=task, default=".", deprecated_keys=["result_path"]))

    separator = get_setting("parameter_separator", task=task, default="=")
    for key, value in serialized_parameters.items():
        # Raise error if parameter value contains path separator "/" ("\" on Windows)
        # or is not interpretable as basename due to other reasons.
        if value != os.path.basename(value):
            raise ValueError(
                f"Parameter `{key}{separator}{value}` cannot be interpreted as directory name. "
                f"Make sure it does not contain the path separators ``{os.path.sep}``. "
                "Consider using a hashed parameter (e.g. ``b2luigi.Parameter(hashed=True)``)."
            )

    use_parameter_name = get_setting("use_parameter_name_in_output", task=task, default=True)
    if use_parameter_name:
        param_list: List[str] = [f"{key}{separator}{value}" for key, value in serialized_parameters.items()]
    else:
        param_list: List[str] = [f"{value}" for value in serialized_parameters.values()]
    output_path: str = os.path.join(result_dir, *param_list)

    return os.path.join(output_path, base_filename)


def get_log_file_dir(task):
    """
    Determines the directory where log files for a given task should be stored.

    If the task has a custom method ``get_log_file_dir``, it will use that method
    to retrieve the log file directory. Otherwise, it constructs the log file
    directory path based on the task's settings and family.

    Args:
        task: The task object for which the log file directory is being determined.

    Returns:
        str: The path to the log file directory for the given task.
    """
    if hasattr(task, "get_log_file_dir"):
        log_file_dir = task.get_log_file_dir()
        return log_file_dir

    # Be sure to evaluate things relative to the current executed file, not to where we are now
    base_log_file_dir = map_folder(get_setting("log_dir", task=task, default="logs", deprecated_keys=["log_folder"]))
    log_file_dir = create_output_file_name(task, task.get_task_family() + "/", result_dir=base_log_file_dir)

    return log_file_dir


def get_task_file_dir(task):
    """
    Determines the directory path for a given task's output files.

    If the task has a method ``get_task_file_dir``, it will use that method
    to retrieve the directory path. Otherwise, it generates the directory
    path using the :obj:`create_output_file_name` function and the task's family name.

    Args:
        task: An object representing the task. It is expected to have a
              `get_task_file_dir` method or a `get_task_family` method.

    Returns:
        str: The directory path for the task's output files.
    """
    if hasattr(task, "get_task_file_dir"):
        task_file_dir = task.get_task_file_dir()
        return task_file_dir

    base_task_file_dir = map_folder(get_setting("task_file_dir", task=task, default="task_files"))
    task_file_dir = create_output_file_name(
        task, task.get_task_family() + "/", result_dir=base_task_file_dir, need_hidden=True
    )

    return task_file_dir


def get_filename():
    """
    Retrieves the absolute path of the main script being executed.

    Returns:
        str: The absolute path of the main script.
    """
    import __main__

    return os.path.abspath(__main__.__file__)


def map_folder(input_folder):
    """
    Maps a relative folder path to an absolute path based on the location of the current script.

    If the input folder path is already absolute, it is returned as-is. Otherwise, the function
    determines the directory of the current script and joins it with the relative input folder path
    to produce an absolute path.

    Args:
        input_folder (str): The folder path to map. Can be either absolute or relative.

    Returns:
        str: The absolute path corresponding to the input folder.

    Raises:
        AttributeError: If the current script location cannot be determined, typically when running
                        in an interactive shell (e.g., Jupyter Notebook). In such cases, the user
                        is advised to provide absolute paths in their settings.
    """
    if os.path.isabs(input_folder):
        return input_folder

    try:
        filename = get_filename()
    except AttributeError as ex:
        raise type(ex)(
            "Could not determine the current script location. "
            "If you are running in an interactive shell (such as jupyter notebook) "
            "make sure to only provide absolute paths in your settings. "
            f"More Info:\n{ex}"
        ).with_traceback(sys.exc_info()[2])

    filepath = os.path.dirname(filename)

    return os.path.join(filepath, input_folder)


def _to_dict(d) -> Dict:
    """
    Converts the input into a dictionary. If the input is already a dictionary, it is returned as-is.
    Otherwise, a new dictionary is created with the input as both the key and the value.

    Args:
        d: The input to be converted into a dictionary. Can be of any type.

    Returns:
        Dict: A dictionary representation of the input.
    """
    if isinstance(d, dict):
        return d

    return {d: d}


def _flatten(struct: Iterable) -> List:
    """
    Recursively flattens a nested iterable structure into a single list.

    Args:
        struct (Iterable): The input structure to flatten.

    Returns:
        List: A flattened list containing all elements from the input structure.

    Notes:
        - If the input is a dictionary or a string, it is returned as a single-element list.
        - Non-iterable inputs are also returned as a single-element list.
    """
    if isinstance(struct, dict) or isinstance(struct, str):
        return [struct]

    result = []
    try:
        iterator = iter(struct)
    except TypeError:
        return [struct]

    for f in iterator:
        result += _flatten(f)

    return result


def on_failure(task, _):
    """
    Handles the failure of a task by generating an explanation message, printing it to stdout,
    and returning it to be sent back to the scheduler.

    Args:
        task: The task instance that failed.

    Returns:
        str: A detailed explanation of the failure, including the task ID, parameters, and
        the location of the log files.
    """
    explanation = f"Failed task {task} with task_id and parameters:\n"
    explanation += f"\ttask_id={task.task_id}\n"
    for key, value in get_filled_params(task).items():
        explanation += f"\t{key}={value}\n"
    explanation += "Please have a look into the log files in:\n"
    explanation += os.path.abspath(get_log_file_dir(task))

    # First print the explanation on stdout
    print(colorama.Fore.RED)
    print(explanation)
    print(colorama.Style.RESET_ALL)

    # Then return it: it will be sent back to the scheduler
    return explanation


def add_on_failure_function(task):
    """
    Assigns a custom failure handler to the given task.

    This function dynamically binds the :obj:`on_failure` method to the provided
    task object.

    Args:
        task: The task object to which the :obj:`on_failure` method will be attached.
    """
    task.on_failure = types.MethodType(on_failure, task)


def create_cmd_from_task(task):
    """
    Constructs a command-line argument list based on the provided task and its settings.

    The executable string is made up of three key components::

        <executable_prefix> <executable> <filename> --batch-runner --task-id ExampleTask_id_123 <task_cmd_additional_args>


    Args:
        task: An object representing the task for which the command is being created.

    Returns:
        list: A list of strings representing the command-line arguments.

    Raises:
        ValueError: If any of the following conditions are met:
            - The ``task_cmd_additional_args`` setting is not a list of strings.
            - The ``executable_prefix`` setting is not a list of strings.
            - The ``executable`` setting is not a list of strings.

    Notes:
        - The ``filename`` is included in the command if the ``add_filename_to_cmd``
          setting is enabled. (Default: ``True``)
        -
    """
    filename = os.path.basename(get_filename()) if get_setting("add_filename_to_cmd", task=task, default=True) else ""
    task_cmd_additional_args = get_setting("task_cmd_additional_args", task=task, default=[])

    if isinstance(task_cmd_additional_args, str):
        raise ValueError("Your specified task_cmd_additional_args needs to be a list of strings, e.g. ['--foo', 'bar']")

    prefix = get_setting("executable_prefix", task=task, default=[], deprecated_keys=["cmd_prefix"])

    if isinstance(prefix, str):
        raise ValueError("Your specified executable_prefix needs to be a list of strings, e.g. [strace]")

    cmd = prefix

    executable = get_setting("executable", task=task, default=[sys.executable])

    if isinstance(executable, str):
        raise ValueError("Your specified executable needs to be a list of strings, e.g. [python3]")

    cmd += executable
    cmd += [filename, "--batch-runner", "--task-id", task.task_id]
    cmd += task_cmd_additional_args

    return cmd


def create_output_dirs(task):
    """
    Creates the necessary output directories for a given task.

    This function takes a task object, retrieves its outputs, and ensures that
    the directories required for those outputs exist by calling the ``makedirs``
    method on each target.

    Args:
        task: The task object whose outputs need directories to be created.
    """
    output_list = flatten_to_dict(task.output())
    output_list = output_list.values()

    for output in output_list:
        output.makedirs()


def get_filled_params(task):
    """
    Retrieve a dictionary of parameter names and their corresponding values
    from a given task.

    Args:
        task: An object representing a task, which must have a `get_params`
              method that returns an iterable of parameter name and metadata
              pairs, and attributes corresponding to the parameter names.

    Returns:
        dict: A dictionary containing parameter names as keys and their
              respective values as values.
    """
    return {key: getattr(task, key) for key, _ in task.get_params()}


def is_subdir(path, parent_dir):
    """
    Determines if a given path is a subdirectory of a specified parent directory.

    Args:
        path (str): The path to check.
        parent_dir (str): The parent directory to compare against.

    Returns:
        bool: ``True`` if the path is a subdirectory of the parent directory, ``False`` otherwise.

    Note:
        Both ``path`` and ``parent_dir`` are converted to their absolute paths before comparison.
    """
    path = os.path.abspath(path)
    parent_dir = os.path.abspath(parent_dir)
    return os.path.commonpath([path, parent_dir]) == parent_dir


def get_apptainer_or_singularity(task=None):
    """
    Determines the command to use for containerization, prioritizing Apptainer
    over Singularity. If neither is available, raises a ValueError.

    The function first checks if a custom command is set via the ``apptainer_cmd``
    setting. If not, it checks for the availability of the ``apptainer`` or
    ``singularity`` commands in the system's PATH.

    Args:
        task (optional): An optional task object that may be used to retrieve
                         task-specific settings.

    Returns:
        str: The command to use for containerization, either "apptainer" or
             "singularity".

    Raises:
        ValueError: If neither Apptainer nor Singularity is available on the
                    system and no custom command is set.
    """
    set_cmd = get_setting("apptainer_cmd", default="", task=task)
    if set_cmd:
        return set_cmd
    if shutil.which("apptainer"):
        return "apptainer"
    elif shutil.which("singularity"):
        return "singularity"
    raise ValueError(
        "Neither apptainer nor singularity is available on this system. If you know that one of them is available on the batch system you can manually set it via the `apptainer_cmd` setting."
    )


def create_apptainer_command(command, task=None):
    """
    Constructs a command to execute within an Apptainer (or Singularity) container.

    This function generates a command string that sets up the necessary environment
    and mounts for running a given command inside an Apptainer container. It ensures
    that required settings are provided and validates compatibility with the batch
    system.

    Args:
        command (str): The command to be executed inside the Apptainer container.
        task (optional): An optional task object or identifier used to retrieve
            task-specific settings.

    Returns:
        list: A list of command-line arguments representing the full Apptainer
        execution command.

    Raises:
        ValueError: If the ``env_script`` is not provided.
        ValueError: If the batch system is ``gbasf2``, as Apptainer is not supported
            for this batch system.

    Notes:
        - ``apptainer_image`` is retrieved from the task settings.
        - ``apptainer_additional_params`` is used to specify additional parameters
          for the Apptainer command. Expecting a string.
        - ``apptainer_mounts`` is used to specify additional mount points for the
          Apptainer command. Expecting a list of strings.
        - ``apptainer_mount_defaults`` determines whether to include default
          mount points (e.g., result directory and log file directory).
    """
    env_setup_script = get_setting("env_script", task=task, default="")
    if not env_setup_script:
        raise ValueError("Apptainer execution requires an environment setup script.")

    apptainer_image = get_setting("apptainer_image", task=task, default="")

    # If the batch system is gbasf2, we cannot use apptainer
    if get_setting("batch_system", default="auto", task=task) == "gbasf2":
        raise ValueError("Invalid batch system for apptainer usage. Apptainer is not supported for gbasf2.")

    exec_command = [get_apptainer_or_singularity(task=task), "exec"]
    additional_params = get_setting("apptainer_additional_params", default="", task=task)
    exec_command += [f" {additional_params}"] if additional_params else []

    # Add apptainer mount points if given
    apptainer_mounts = get_setting("apptainer_mounts", task=task, default=None)
    if apptainer_mounts is None:
        mounts = []
    else:
        mounts = copy.copy(apptainer_mounts)

    if get_setting("apptainer_mount_defaults", task=task, default=True):
        local_mounts = [
            map_folder(get_setting("result_dir", task=task, default=".", deprecated_keys=["result_path"])),
            get_log_file_dir(task=task),
        ]
        for mount in local_mounts:
            os.makedirs(mount, exist_ok=True)
            if not is_subdir(mount, os.getcwd()):
                mounts.append(mount)

    for mount in mounts:
        exec_command += ["--bind", mount]

    exec_command += [apptainer_image]
    exec_command += ["/bin/bash", "-c"]
    exec_command += [f"'source {env_setup_script} && {command}'"]

    # Do the shlex split for correct string interpretation
    return shlex.split(" ".join(exec_command))


def get_luigi_logger():
    """
    Retrieve the logger instance used by ``luigi``.

    Returns:
        logging.Logger: The logger instance for "luigi-interface".
    """
    logger = logging.getLogger("luigi-interface")
    return logger



"""
.. _exercise05_label:

Introduction to ``Basf2PathTask``
=================================

.. hint::
    This example demonstrates how to create a simple b2luigi task that executes a
    ``basf2`` simulation.
    The key points to be learned from this example are the
    :class:`Basf2PathTask <b2luigi.basf2_helper.tasks.Basf2PathTask>` class
    and :meth:`create_path <b2luigi.basf2_helper.tasks.Basf2PathTask.create_path>` method.

For directly working with the Belle II Analysis Software Framework (``basf2``),
``b2luigi`` provides the :class:`Basf2PathTask <b2luigi.basf2_helper.tasks.Basf2PathTask>`
task class.
This class provides help for the user to create and execute a ``basf2`` path.
The :class:`Basf2PathTask <b2luigi.basf2_helper.tasks.Basf2PathTask>` also provides the
following parameters that can be useful when executing ``basf2`` processes:

- ``num_processes``: number of parallel processes to use for the ``basf2`` execution.

- ``max_event``: maximum number of events to process. If set to ``0``, all events
  will be processed.

In contrast to the normal ``(b2)luigi`` tasks, the execution logic of a
:class:`Basf2PathTask <b2luigi.basf2_helper.tasks.Basf2PathTask>` is not defined in
a ``run`` method but in
:meth:`create_path <b2luigi.basf2_helper.tasks.Basf2PathTask.create_path>`.
The :meth:`create_path <b2luigi.basf2_helper.tasks.Basf2PathTask.create_path>` method needs
to return the ``basf2`` path that is created in the steering file.
Furthermore, the ``Progress`` module is automatically added and ``print(b2.statistics)``
is called after the path is processed.

.. warning::
    Due to technical reasons, the path needs to be created within the :meth:`create_path <b2luigi.basf2_helper.tasks.Basf2PathTask.create_path>`
    method. The path can be used in further objects, however, it is not possible
    for it to originate from an outer scope.

Additional tips:

- For later purposes, we add a parameter (``identifier``) to identify the output file
  uniqley. This is useful when running multiple tasks in parallel.

- We make use of the task own settings via `result_dir`. It is sometimes more useful
  to structure the output of your tasks in task specific directories than solely in the
  parameter based directories. Additionally, good practice is to provide absolute paths
  instead of relative paths as we do in this example.

- No batch submission of this task will be done (``batch_system = "local"``), more will
  come in subsequent examples.

- The example relies again on the task's own
  :meth:`get_output_file_name <b2luigi.Task.get_output_file_name>` method to generate
  the output file name.
"""

import basf2 as b2
import generators as ge
import simulation as si

import b2luigi
from b2luigi.basf2_helper import Basf2PathTask


class SimulationTask(Basf2PathTask):
    n_events = b2luigi.IntParameter()
    event_type = b2luigi.Parameter(default="mumu")
    identifier = b2luigi.Parameter(default="sample")

    result_dir = "results/Simulation"
    batch_system = "local"

    def output(self):
        yield self.add_to_output("simulation.root")

    def create_path(self):
        main = b2.Path()

        # Set number of events (optional: also set experiment and run number)
        main.add_module("EventInfoSetter", evtNumList=[self.n_events])
        decfile = (
            "./decfiles/BptoKpJpsi_JpsitoMuMu.dec"
            if str(self.event_type) == "mumu"
            else "./decfiles/BptoKpJpsi_JpsitoEE.dec"
        )
        # Signal generation
        ge.add_evtgen_generator(
            path=main, finalstate="signal", signaldecfile=b2.find_file(decfile), eventType=self.event_type
        )

        # Simulation
        si.add_simulation(path=main)

        # Write out simulation output to root file.
        main.add_module("RootOutput", outputFileName=self.get_output_file_name("simulation.root"))

        return main


if __name__ == "__main__":
    b2luigi.process(SimulationTask(n_events=10))



from b2luigi.cli.arguments import get_cli_arguments
from b2luigi.cli import runner


__has_run_already = False


def process(
    task_like_elements,
    show_output=False,
    dry_run=False,
    test=False,
    batch=False,
    remove=[],
    remove_only=[],
    auto_confirm=False,
    keep_tasks=None,
    ignore_additional_command_line_args=False,
    **kwargs,
):
    """
    Call this function in your main method to tell ``b2luigi`` where your entry
    point of the task graph is.
    It is very similar to ``luigi.build`` with some additional configuration options.

    Example:
        This example defines a simple task and tells ``b2luigi`` to execute it 100 times
        with different parameters:

        .. code-block:: python

          import b2luigi
          import random

          class MyNumberTask(b2luigi.Task):
              some_parameter = b2luigi.Parameter()

              def output(self):
                  return b2luigi.LocalTarget(f"results/output_file_{self.some_parameter}.txt")

              def run(self):
                  random_number = random.random()
                  with self.output().open("w") as f:
                      f.write(f"{random_number}\\n")

          if __name__ == "__main__":
              b2luigi.process([MyNumberTask(some_parameter=i) for i in range(100)])

    All flag arguments can also be given as command line arguments.
    This means the call with::

        b2luigi.process(tasks, batch=True)

    is equivalent to calling the script with::

        python script.py --batch

    Args:
        task_like_elements (:obj:`Task` or list): Task(s) to execute with luigi.
            Can either be a list of tasks or a task instance.

        show_output (bool, optional): Instead of running the task(s), write out all output files
            which will be generated marked in color, if they are present already.
            Good for testing of your tasks will do, what you think they should.

        dry_run (bool, optional): Instead od running the task(s), write out which tasks will
            be executed. This is a simplified form of dependency resolution, so this
            information may be wrong in some corner cases. Also good for testing.

        test (bool, optional): Does neither run on the batch system, with multiprocessing
            or dispatched (see :obj:`DispatchableTask`) but directly on the machine for
            debugging reasons. Does output all logs to the console.

        batch (bool, optional): If set to `False`, the global settings of `batch_system`
            will be set to `local`. If set to `True`, task with no batch_system set,
            will be executed with the globally set `batch_system`.
            Refer to :ref:`quick-start-label` for more information.
            By default, the global batch system uses the `auto` setting, but this can be changed with the
            `batch_system` settings. See :meth:`get_setting <b2luigi.core.settings.get_setting>` on how to define settings.

        remove (list, optional): If a single task is given, remove the output of this task.
            If a list of tasks is given, remove the output of all tasks in the list.

        ignore_additional_command_line_args (bool, optional, default False): Ignore additional
            command line arguments. This is useful if you want to use this function in a file
            that also does some command line parsing.

        **kwargs: Additional keyword arguments passed to ``luigi.build``.

    Warning:
        You should always have just a single call to ``process`` in your script.
        If you need to have multiple calls, either use a :class:`b2luigi.WrapperTask`
        or two scripts.

    """
    # Assert, that process is only run once
    global __has_run_already
    if __has_run_already:
        raise RuntimeError("You are not allowed to call process twice in your code!")
    __has_run_already = True

    # Create Task List
    if not isinstance(task_like_elements, list):
        task_list = [task_like_elements]
    else:
        task_list = task_like_elements

    # Check the CLI arguments and run as requested
    cli_args = get_cli_arguments(ignore_additional_command_line_args=ignore_additional_command_line_args)

    if cli_args.show_output or show_output:
        runner.show_all_outputs(task_list)
    elif cli_args.dry_run or dry_run:
        runner.dry_run(task_list)
    elif cli_args.test or test:
        runner.run_test_mode(task_list, cli_args, kwargs)
    elif cli_args.batch_runner:
        runner.run_as_batch_worker(task_list, cli_args, kwargs)
    elif cli_args.remove or remove:
        runner.remove_outputs(
            task_list,
            target_tasks=cli_args.remove or remove,
            auto_confirm=auto_confirm or cli_args.yes,
            keep_task=cli_args.keep or keep_tasks,
        )
    elif cli_args.remove_only or remove_only:
        runner.remove_outputs(
            task_list,
            target_tasks=cli_args.remove_only or remove_only,
            only=True,
            auto_confirm=auto_confirm or cli_args.yes,
            keep_task=cli_args.keep or keep_tasks,
        )
    elif cli_args.batch or batch:
        runner.run_batched(task_list, cli_args, kwargs)
    else:
        runner.run_local(task_list, cli_args, kwargs)



import os
from unittest import TestCase, mock

import b2luigi
import shlex
from b2luigi.core import utils
from ..helpers import B2LuigiTestCase


class ProductDictTestCase(TestCase):
    def test_basic_usage(self):
        kwargs = list(utils.product_dict(first_arg=[1, 2, 3], second_arg=["a", "b"]))

        self.assertEqual(len(kwargs), 6)

        self.assertIn({"first_arg": 1, "second_arg": "a"}, kwargs)
        self.assertIn({"first_arg": 1, "second_arg": "b"}, kwargs)
        self.assertIn({"first_arg": 2, "second_arg": "a"}, kwargs)
        self.assertIn({"first_arg": 2, "second_arg": "b"}, kwargs)
        self.assertIn({"first_arg": 3, "second_arg": "a"}, kwargs)
        self.assertIn({"first_arg": 3, "second_arg": "b"}, kwargs)

        kwargs = list(utils.product_dict(first_arg=[1, 2, 3]))

        self.assertEqual(len(kwargs), 3)

        self.assertIn({"first_arg": 1}, kwargs)
        self.assertIn({"first_arg": 2}, kwargs)
        self.assertIn({"first_arg": 3}, kwargs)

        kwargs = list(utils.product_dict(first_arg=[1, 2, 3], second_arg=[]))

        self.assertEqual(len(kwargs), 0)

        kwargs = list(utils.product_dict(first_arg=[1], second_arg=["a"]))

        self.assertEqual(len(kwargs), 1)

        self.assertIn({"first_arg": 1, "second_arg": "a"}, kwargs)


class FlattenTestCase(TestCase):
    def test_list_input(self):
        inputs = [{"key1": "value1"}, {"key2": "value2"}]

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], "value1")
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], "value2")

        inputs = [{"key1": "value1"}, {"key2": "value2"}, {"key1": "repeated"}]

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], "repeated")
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], "value2")

        inputs = []

        outputs = utils.flatten_to_dict(inputs)

        self.assertFalse(outputs)

        inputs = [{"key1": "value1"}, "value2"]

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], "value1")
        self.assertIn("value2", outputs)
        self.assertEqual(outputs["value2"], "value2")

        inputs = ["value1", "value2"]

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("value1", outputs)
        self.assertEqual(outputs["value1"], "value1")
        self.assertIn("value2", outputs)
        self.assertEqual(outputs["value2"], "value2")

    def test_raw_input(self):
        inputs = {"key1": "value1"}

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], "value1")

        inputs = "value1"

        outputs = utils.flatten_to_dict(inputs)

        self.assertIn("value1", outputs)
        self.assertEqual(outputs["value1"], "value1")

    def test_list_of_list_input(self):
        inputs = [{"key1": "value1"}, {"key2": "value2"}]

        outputs = utils.flatten_to_dict_of_lists(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], ["value1"])
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], ["value2"])

        inputs = [{"key1": "value1"}, {"key2": "value2"}, {"key1": "repeated"}]

        outputs = utils.flatten_to_dict_of_lists(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], ["value1", "repeated"])
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], ["value2"])

        inputs = []

        outputs = utils.flatten_to_dict_of_lists(inputs)

        self.assertFalse(outputs)

        inputs = [{"key1": "value1"}, "value2"]

        outputs = utils.flatten_to_dict_of_lists(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], ["value1"])
        self.assertIn("value2", outputs)
        self.assertEqual(outputs["value2"], ["value2"])

        inputs = ["value1", "value2"]

        outputs = utils.flatten_to_dict_of_lists(inputs)

        self.assertIn("value1", outputs)
        self.assertEqual(outputs["value1"], ["value1"])
        self.assertIn("value2", outputs)
        self.assertEqual(outputs["value2"], ["value2"])

        inputs = [[{"key1": "value1"}, {"key2": "value2"}], [{"key1": "repeated"}]]

        outputs = utils.flatten_to_dict_of_lists(inputs)

        self.assertIn("key1", outputs)
        self.assertEqual(outputs["key1"], ["value1", "repeated"])
        self.assertIn("key2", outputs)
        self.assertEqual(outputs["key2"], ["value2"])


class OutputFileNameTestCase(B2LuigiTestCase):
    dummy_dir = "/foo/bar/"
    dummy_filename = "/foo/bar/run.py"
    dummy_parameter = "foo"

    def _get_dummy_task(self, parameter):
        class MyTask(b2luigi.Task):
            parameter = b2luigi.Parameter()

        task = MyTask(parameter)
        return task

    def test_output_file_name_basename(self):
        """
        Test that utils.create_output_file_name will return the expected output filename path
        """
        test_task = self._get_dummy_task(self.dummy_parameter)
        output = utils.create_output_file_name(
            test_task, base_filename="output.txt", result_dir=f"{self.test_dir}/results"
        )

        self.assertEqual(output, f"{self.test_dir}/results/parameter={self.dummy_parameter}/output.txt")

    def test_output_file_name_basename_alternate_separator(self):
        """
        Test that utils.create_output_file_name will return the expected output filename path
        """
        b2luigi.set_setting("parameter_separator", "_EQ_")

        test_task = self._get_dummy_task(self.dummy_parameter)
        output = utils.create_output_file_name(
            test_task, base_filename="output.txt", result_dir=f"{self.test_dir}/results"
        )
        b2luigi.clear_setting("parameter_separator")
        self.assertEqual(output, f"{self.test_dir}/results/parameter_EQ_{self.dummy_parameter}/output.txt")

    def test_output_file_name_basename_error(self):
        """
        Test that utils.create_output_file_name will raise an error if a parameter contains
        path separator "/" or is not interpretable as basename due to other reasons.
        """

        test_task_dir = self._get_dummy_task(self.dummy_dir)
        with self.assertRaises(ValueError):
            utils.create_output_file_name(test_task_dir, base_filename="output.txt")

        test_task_filename = self._get_dummy_task(self.dummy_filename)
        with self.assertRaises(ValueError):
            utils.create_output_file_name(test_task_filename, base_filename="output.txt")


class MapFolderTestCase(TestCase):
    dummy_rel_dir = "./some/rel_dir"
    dummy_abs_dir = "/path/to/some/abs_dir"
    main_no_file_file_err_msg = "module '__main__' has no attribute '__file__'"

    def test_map_folder_abspath_identity(self):
        """Test that for an absolute path, map_folder returns and identity"""
        self.assertEqual(utils.map_folder(self.dummy_abs_dir), self.dummy_abs_dir)

    def test_map_folder_relpath(self):
        """
        Test map_folder with a relative input_folder, which joins it with ``__main__.__file__``
        """
        with mock.patch("__main__.__file__", self.dummy_abs_dir):
            mapped_folder = utils.map_folder(self.dummy_rel_dir)
            self.assertEqual(mapped_folder, os.path.join(self.dummy_abs_dir, mapped_folder))

    def test_map_folder_abspath_identity_when_no_filename(self):
        """
        Test that for an absolute path, map_folder returns and identity even
        if ``get_filename`` would raise an ``AttributeError`` because ``__main__.__file__``
        is not available (e.g. in jupyter)
        """
        with mock.patch(
            "b2luigi.core.utils.get_filename",
            side_effect=AttributeError(self.main_no_file_file_err_msg),
        ):
            mapped_folder = utils.map_folder(self.dummy_abs_dir)
            self.assertEqual(mapped_folder, self.dummy_abs_dir)

    def test_map_folder_raises_attribute_error_for_relpath_when_no_filename(self):
        """
        Test that when ``get_filename`` returns an ``AttributeError`` b/c
        ``__main__.__file__`` is not available, ``map_folder`` also returns an
        ``AttributeError`` when the input folder is relative
        """
        with self.assertRaises(AttributeError):
            with mock.patch(
                "b2luigi.core.utils.get_filename",
                side_effect=AttributeError(self.main_no_file_file_err_msg),
            ):
                utils.map_folder(self.dummy_rel_dir)

    def _get_map_folder_error_mesage(self):
        """
        Get the error message that ``map_folder`` raises when ``get_filename``
        raises an attribute error because ``__main__.__file__`` is not
        accessible (e.g. in Jupyter)
        """
        try:
            with mock.patch(
                "b2luigi.core.utils.get_filename",
                side_effect=AttributeError(self.main_no_file_file_err_msg),
            ):
                utils.map_folder(self.dummy_rel_dir)
        except AttributeError as err:
            return str(err)
        raise RuntimeError("No AttributeError raised when calling ``utils.map_folder``")

    def test_original_message_in_error(self):
        """
        Check that the error message of ``map_folder`` still contains the
        original error message raised by ``get_filename`` during an
        ``AttributeError`` due to ``__main__.__file__`` not being accessible
        """
        message = self._get_map_folder_error_mesage()
        self.assertTrue(message.endswith(self.main_no_file_file_err_msg))

    def test_additional_info_added_to_error(self):
        """
        Check that the error message of ``map_folder`` adds additional
        information to the ``AttributeError`` raised by ``get_filename``
        """
        message = self._get_map_folder_error_mesage()
        self.assertTrue(message.startswith("Could not determine the current script location."))


class TaskIteratorTestCase(TestCase):
    def test_task_iterator_unique_tasks(self):
        """
        Test that even when multiple worker tasks require same common
        dependency task, it appears only once in task iterator output.
        """

        class CommonDependencyTask(b2luigi.ExternalTask):
            def output(self):
                return b2luigi.LocalTarget("some_dependency")

        @b2luigi.requires(CommonDependencyTask)
        class WorkerTask(b2luigi.Task):
            some_parameter = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("output")

        class AggregatorTask(b2luigi.WrapperTask):
            def requires(self):
                for param in range(3):
                    yield self.clone(WorkerTask, some_parameter=param)

        expected_task_str_order = [
            "AggregatorTask()",
            "WorkerTask(some_parameter=0)",
            "CommonDependencyTask()",
            "WorkerTask(some_parameter=1)",
            "WorkerTask(some_parameter=2)",
        ]
        resulting_task_str_order = [str(t) for t in utils.task_iterator(AggregatorTask())]
        self.assertListEqual(expected_task_str_order, resulting_task_str_order)


class IsSubdirTestCase(TestCase):
    def test_is_subdir_true(self):
        self.assertTrue(utils.is_subdir("/path/to/child", "/path/to"))
        self.assertTrue(utils.is_subdir("/path/to/child/grandchild", "/path/to"))
        self.assertTrue(utils.is_subdir("/path/to/child", "/path/to/child"))
        self.assertTrue(utils.is_subdir("/path/to/child/", "/path/to"))
        self.assertTrue(utils.is_subdir("/path/to/child", "/path/to/"))

    def test_is_subdir_false(self):
        self.assertFalse(utils.is_subdir("/path/to/child", "/path/to/other"))
        self.assertFalse(utils.is_subdir("/path/to/child", "/path/to/child/grandchild"))
        self.assertFalse(utils.is_subdir("/path/to", "/path/to/child"))
        self.assertFalse(utils.is_subdir("/path/to/child", "/other/path/to"))

    def test_is_subdir_relative_paths(self):
        self.assertTrue(utils.is_subdir("child", "."))
        self.assertTrue(utils.is_subdir("child/grandchild", "."))
        self.assertFalse(utils.is_subdir(".", "child"))
        self.assertFalse(utils.is_subdir("child", "other"))


class CreateApptainerCommandTestCase(TestCase):
    def setUp(self):
        self.command = "echo Hello World"
        self.task = mock.Mock()
        self.env_setup_script = "/path/to/env_setup.sh"
        self.apptainer_image = "/path/to/apptainer_image.sif"
        self.result_dir = "/path/to/results"
        self.log_dir = "/path/to/logs"
        self.additional_params = "--nv"
        self.mounts = ["/mnt/data"]

        self.settings = {
            "env_script": self.env_setup_script,
            "apptainer_image": self.apptainer_image,
            "result_dir": self.result_dir,
            "log_dir": self.log_dir,
            "apptainer_additional_params": self.additional_params,
            "apptainer_mounts": self.mounts,
            "apptainer_mount_defaults": True,
            "batch_system": "lsf",
        }

    def test_create_apptainer_command(self):
        with mock.patch(
            "b2luigi.core.utils.get_setting",
            side_effect=lambda key, **kwargs: self.settings.get(key, kwargs.get("default")),
        ):
            with mock.patch("b2luigi.core.utils.map_folder", side_effect=lambda x: x):
                with mock.patch("b2luigi.core.utils.get_log_file_dir", return_value=self.log_dir):
                    with mock.patch("os.makedirs"):
                        expected_command = [
                            "apptainer",
                            "exec",
                            f" {self.additional_params}",
                            "--bind",
                            self.mounts[0],
                            "--bind",
                            self.result_dir,
                            "--bind",
                            self.log_dir,
                            self.apptainer_image,
                            "/bin/bash",
                            "-c",
                            f"'source {self.env_setup_script} && {self.command}'",
                        ]
                        with mock.patch("b2luigi.core.utils.get_apptainer_or_singularity", return_value="apptainer"):
                            result = utils.create_apptainer_command(self.command, task=self.task)
                        self.assertEqual(result, shlex.split(" ".join(expected_command)))

    def test_create_apptainer_command_no_env_script(self):
        self.settings["env_script"] = ""
        with mock.patch(
            "b2luigi.core.utils.get_setting",
            side_effect=lambda key, **kwargs: self.settings.get(key, kwargs.get("default")),
        ):
            with self.assertRaises(ValueError) as context:
                utils.create_apptainer_command(self.command, task=self.task)
            self.assertEqual(str(context.exception), "Apptainer execution requires an environment setup script.")

    def test_create_apptainer_command_invalid_batch_system(self):
        self.settings["batch_system"] = "gbasf2"
        with mock.patch(
            "b2luigi.core.utils.get_setting",
            side_effect=lambda key, **kwargs: self.settings.get(key, kwargs.get("default")),
        ):
            with self.assertRaises(ValueError) as context:
                utils.create_apptainer_command(self.command, task=self.task)
            self.assertEqual(
                str(context.exception),
                "Invalid batch system for apptainer usage. Apptainer is not supported for gbasf2.",
            )

    def test_create_apptainer_command_no_additional_params(self):
        self.settings["apptainer_additional_params"] = ""
        with mock.patch(
            "b2luigi.core.utils.get_setting",
            side_effect=lambda key, **kwargs: self.settings.get(key, kwargs.get("default")),
        ):
            with mock.patch("b2luigi.core.utils.map_folder", side_effect=lambda x: x):
                with mock.patch("b2luigi.core.utils.get_log_file_dir", return_value=self.log_dir):
                    with mock.patch("os.makedirs"):
                        expected_command = [
                            "apptainer",
                            "exec",
                            "--bind",
                            self.mounts[0],
                            "--bind",
                            self.result_dir,
                            "--bind",
                            self.log_dir,
                            self.apptainer_image,
                            "/bin/bash",
                            "-c",
                            f"'source {self.env_setup_script} && {self.command}'",
                        ]
                        with mock.patch("b2luigi.core.utils.get_apptainer_or_singularity", return_value="apptainer"):
                            result = utils.create_apptainer_command(self.command, task=self.task)
                        self.assertEqual(result, shlex.split(" ".join(expected_command)))

    def test_create_apptainer_command_no_mounts(self):
        self.settings["apptainer_mounts"] = []
        with mock.patch(
            "b2luigi.core.utils.get_setting",
            side_effect=lambda key, **kwargs: self.settings.get(key, kwargs.get("default")),
        ):
            with mock.patch("b2luigi.core.utils.map_folder", side_effect=lambda x: x):
                with mock.patch("b2luigi.core.utils.get_log_file_dir", return_value=self.log_dir):
                    with mock.patch("os.makedirs"):
                        expected_command = [
                            "apptainer",
                            "exec",
                            f" {self.additional_params}",
                            "--bind",
                            self.result_dir,
                            "--bind",
                            self.log_dir,
                            self.apptainer_image,
                            "/bin/bash",
                            "-c",
                            f"'source {self.env_setup_script} && {self.command}'",
                        ]
                        with mock.patch("b2luigi.core.utils.get_apptainer_or_singularity", return_value="apptainer"):
                            result = utils.create_apptainer_command(self.command, task=self.task)
                        self.assertEqual(result, shlex.split(" ".join(expected_command)))

    def test_create_apptainer_command_override_apptainer_name(self):
        with mock.patch(
            "b2luigi.core.utils.get_setting",
            side_effect=lambda key, **kwargs: self.settings.get(key, kwargs.get("default")),
        ):
            with mock.patch("b2luigi.core.utils.map_folder", side_effect=lambda x: x):
                with mock.patch("b2luigi.core.utils.get_log_file_dir", return_value=self.log_dir):
                    with mock.patch("os.makedirs"):
                        expected_command = [
                            "test_apptainer_cmd",
                            "exec",
                            f" {self.additional_params}",
                            "--bind",
                            self.mounts[0],
                            "--bind",
                            self.result_dir,
                            "--bind",
                            self.log_dir,
                            self.apptainer_image,
                            "/bin/bash",
                            "-c",
                            f"'source {self.env_setup_script} && {self.command}'",
                        ]
                        self.settings["apptainer_cmd"] = "test_apptainer_cmd"
                        result = utils.create_apptainer_command(self.command, task=self.task)
                        self.assertEqual(result, shlex.split(" ".join(expected_command)))
                        del self.settings["apptainer_cmd"]



from datetime import datetime
import json
import os
import tempfile
import unittest
from collections import Counter
from typing import List
from unittest.mock import MagicMock, Mock, patch

import b2luigi
from b2luigi.batch.processes.gbasf2 import Gbasf2Process, JobStatus, get_unique_project_name

from ..helpers import B2LuigiTestCase
from .batch_task_1 import MyTask


class MyGbasf2Task(MyTask):
    gbasf2_project_name_prefix = "my_gb2_task_"


class TestGbasf2RescheduleJobs(B2LuigiTestCase):
    job_statuses_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "_gbasf2_project_statuses")
    joblist_tmpfile_name = "jobs_to_be_rescheduled.txt"

    def setUp(self):
        super().setUp()
        self.gb2_mock_process = MagicMock()
        self.gb2_mock_process.task = MyGbasf2Task("some_parameter")
        self.gb2_mock_process.dirac_user = "username"
        self.gb2_mock_process.gbasf2_project_name = get_unique_project_name(self.gb2_mock_process.task)
        self.gb2_mock_process.n_retries_by_job = Counter()
        self.gb2_mock_process.max_retries = 1
        b2luigi.set_setting("gbasf2_print_status_updates", False)

    def _get_job_status_dict(self, job_status_fname):
        with open(os.path.join(self.job_statuses_dir, job_status_fname)) as job_status_json_file:
            return json.load(job_status_json_file)

    def _reschedule_jobs(self, job_ids):
        pass

    def assert_rescheduled_jobs(self, job_status_fname, expected_jobs_to_be_rescheduled):
        with patch(
            "b2luigi.batch.processes.gbasf2.get_gbasf2_project_job_status_dict",
            MagicMock(return_value=self._get_job_status_dict(job_status_fname)),
        ):
            with patch("b2luigi.batch.processes.gbasf2.Gbasf2Process._reschedule_jobs", new=self._reschedule_jobs):
                Gbasf2Process._reschedule_failed_jobs(self.gb2_mock_process)

                self.assertCountEqual(
                    list(self.gb2_mock_process.n_retries_by_job.keys()), expected_jobs_to_be_rescheduled
                )
                for jobid in self.gb2_mock_process.n_retries_by_job.keys():
                    self.assertEqual(self.gb2_mock_process.n_retries_by_job[jobid], 1)

    def test_reschedule_jobs_all_done(self):
        "Test gbasf2 project rescheduling with dict where all jobs are done"
        self.assert_rescheduled_jobs("done_testjbucket1357828d80b3.json", [])

    def test_reschedule_jobs_one_failed(self):
        "Test gbasf2 project rescheduling with dict where one job in project failed"
        self.assert_rescheduled_jobs("failed_7663_r03743_10_prod00013766_11x1.json", ["188623842", "188625261"])

    def test_reschedule_jobs_running(self):
        "Test gbasf2 project rescheduling with dict where several jobs are either running or waiting"
        self.assert_rescheduled_jobs("running_TrainingFEI_17-04-2021a10a957289.json", [])

    def test_reschedule_jobs_major_status_done_but_minor_status_not(self):
        "Test gbasf2 project rescheduling with dict where major job status is all done but application status is not ``DONE``"
        self.assert_rescheduled_jobs("all_done_but_application_error.json", ["187522107"])


class TestGbasf2GetJobStatus(B2LuigiTestCase):
    job_statuses_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "_gbasf2_project_statuses")

    def setUp(self):
        super().setUp()
        self.gb2_mock_process = Mock()
        self.gb2_mock_process.task = MyGbasf2Task("some_parameter")
        self.gb2_mock_process.dirac_user = "username"
        self.gb2_mock_process.gbasf2_project_name = get_unique_project_name(self.gb2_mock_process.task)
        self.gb2_mock_process.max_retries = 0
        self.gb2_mock_process._last_job_status_update = datetime.fromtimestamp(0)
        self.gb2_mock_process._job_status_max_age = 5 * 60
        b2luigi.set_setting("gbasf2_print_status_updates", False)

    def _get_job_status_dict(self, job_status_fname):
        with open(os.path.join(self.job_statuses_dir, job_status_fname)) as job_status_json_file:
            return json.load(job_status_json_file)

    def assert_job_status(self, job_status_fname, expected_job_status):
        with patch(
            "b2luigi.batch.processes.gbasf2.get_gbasf2_project_job_status_dict",
            MagicMock(return_value=self._get_job_status_dict(job_status_fname)),
        ):
            job_status = Gbasf2Process.get_job_status(self.gb2_mock_process)
            self.assertEqual(job_status, expected_job_status)

    def test_get_job_status_all_done(self):
        "Test gbasf2 project status dict where all jobs are done"
        self.assert_job_status("done_testjbucket1357828d80b3.json", JobStatus.successful)

    def test_get_job_status_one_failed(self):
        "Test gbasf2 project status dict where one job in project failed"
        self.assert_job_status("failed_7663_r03743_10_prod00013766_11x1.json", JobStatus.aborted)

    def test_get_job_status_running(self):
        "Test gbasf2 project status dict where several jobs are either running or waiting"
        self.assert_job_status("running_TrainingFEI_17-04-2021a10a957289.json", JobStatus.running)

    def test_get_job_status_major_status_done_but_minor_status_not(self):
        "Test gbasf2 project status dict where major job status is all done but application status is not ``DONE``"
        self.assert_job_status("all_done_but_application_error.json", JobStatus.aborted)


class TestBuildGbasf2SubmitCommand(B2LuigiTestCase):
    dummy_lfn = "/grid/path/to/dataset"

    def setUp(self):
        super().setUp()
        self.dummy_lfn_file = tempfile.NamedTemporaryFile()

    def _get_task_file_dir(self, task):
        return self.test_dir

    def _build_gbasf2_submit_command(self, task) -> List[str]:
        gb2_mock_process = Mock()
        os.makedirs(self._get_task_file_dir(task), exist_ok=True)
        self.pickle_file_path = os.path.join(self._get_task_file_dir(task), "serialized_basf2_path.pkl")
        self.wrapper_file_path = os.path.join(self._get_task_file_dir(task), "steering_file_wrapper.py")
        gb2_mock_process.pickle_file_path = self.test_dir
        gb2_mock_process.wrapper_file_path = self.test_dir
        gb2_mock_process.gbasf2_project_name = get_unique_project_name(task)
        gb2_mock_process.task = task
        return Gbasf2Process._build_gbasf2_submit_command(gb2_mock_process)

    def test_input_dataset_in_command(self):
        "Test that if the input dataset is set, it occurs once in the command string"

        class _Task(MyGbasf2Task):
            gbasf2_input_dataset = b2luigi.Parameter(default=self.dummy_lfn)

        gb2_cmd = self._build_gbasf2_submit_command(_Task("some_parameter"))
        gb2_cmd_str = " ".join(gb2_cmd)
        self.assertEqual(gb2_cmd.count(self.dummy_lfn), 1)
        self.assertEqual(gb2_cmd.count("-i"), 1)
        self.assertEqual(gb2_cmd_str.count(f" -i {self.dummy_lfn}"), 1)
        self.assertNotIn("--input_dslist", gb2_cmd_str)

    def test_input_dslist_in_command(self):
        "Test that if the input dataset list is set, it occurs once in the command string"

        class _Task(MyGbasf2Task):
            gbasf2_input_dslist = self.dummy_lfn_file.name

        gb2_cmd = self._build_gbasf2_submit_command(_Task("some_parameter"))
        gb2_cmd_str = " ".join(gb2_cmd)
        self.assertEqual(gb2_cmd.count(self.dummy_lfn_file.name), 1)
        self.assertEqual(gb2_cmd.count("--input_dslist"), 1)
        self.assertEqual(gb2_cmd_str.count(f" --input_dslist {self.dummy_lfn_file.name}"), 1)
        self.assertNotIn("-i", gb2_cmd)

    def test_input_dslist_set_but_does_not_exist(self):
        """Test that if the input dataset list is set but the file does not exist, an ``FileNotFoundError`` is raised"""

        class _Task(MyGbasf2Task):
            gbasf2_input_dslist = self.dummy_lfn_file.name

        self.dummy_lfn_file.close()
        self.assertRaises(FileNotFoundError, lambda: self._build_gbasf2_submit_command(_Task("some_parameter")))

    def test_not_dataset_raises_error(self):
        """Test that we get an error when no gbasf2 input dataset is provided"""
        self.assertRaises(RuntimeError, lambda: self._build_gbasf2_submit_command(MyGbasf2Task("some_parameter")))

    def test_both_input_dataset_and_dslist_raises_error(self):
        """Test that en error is raise when both gbasf2_input_dataset and gbasf2_inputdslist settings are given"""

        class _Task(MyGbasf2Task):
            gbasf2_input_dataset = b2luigi.Parameter(default=self.dummy_lfn)
            gbasf2_input_dslist = self.dummy_lfn_file.name

        self.assertRaises(RuntimeError, lambda: self._build_gbasf2_submit_command(MyGbasf2Task("some_parameter")))


class TestGetGbasf2DatasetQuery(unittest.TestCase):
    "Run tests for ``Gbasf2Process._get_gbasf2_dataset_query``."

    def setUp(self):
        super().setUp()
        self.gb2_mock_process = Mock()
        self.gb2_mock_process.gbasf2_project_name = "projectname"
        self.gb2_mock_process.dirac_user = "username"
        self.gb2_mock_process.task = MyTask("some_parameter")

    def test_filename_single_root_extension(self):
        dataset_query = Gbasf2Process._get_gbasf2_dataset_query(self.gb2_mock_process, output_file_name="output.root")
        self.assertEqual(dataset_query, "/belle/user/username/projectname/sub*/output_*.root")

    def test_filename_multi_extension(self):
        dataset_query = Gbasf2Process._get_gbasf2_dataset_query(
            self.gb2_mock_process, output_file_name="output.mdst.root"
        )
        self.assertEqual(dataset_query, "/belle/user/username/projectname/sub*/output_*.mdst.root")

    def test_non_root_extension_raises_err(self):
        "Raise an error if gbasf2 output does not end with .root"
        with self.assertRaises(ValueError):
            Gbasf2Process._get_gbasf2_dataset_query(self.gb2_mock_process, output_file_name="output.mdst")

    def test_non_basename_raises_err(self):
        """
        Currently we only support simple basenames as gbasf2 job outputs.
        LPN is guessed via dirac username for user projects.
        """
        with self.assertRaises(ValueError):
            Gbasf2Process._get_gbasf2_dataset_query(self.gb2_mock_process, output_file_name="/path/to/output.root")

    def test_filename_single_extension(self):
        gb2_mock_process = Mock()
        gb2_mock_process.gbasf2_project_name = "projectname"
        gb2_mock_process.dirac_user = "username"
        gb2_mock_process.task = MyTask("some_parameter")

        dataset_query = Gbasf2Process._get_gbasf2_dataset_query(self=gb2_mock_process, output_file_name="output.root")
        self.assertEqual(dataset_query, "/belle/user/username/projectname/sub*/output_*.root")



"""
.. _exercise06_label:

The ``requires`` decorator
==========================

.. hint::
    This example demonstrates how to create a simple ``b2luigi`` task that
    executes the reconstruction of previously simulated events. The key point of
    this example is the ``requires`` decorator.

We start with the usual imports. Note that we import the task ``SimulationTask``
class defined in the previous example :ref:`exercise05_label`: it's generally recommended
to keep your code as much modular as possible.
"""

import basf2 as b2
import reconstruction as re
import mdst

import b2luigi
from b2luigi.basf2_helper import Basf2PathTask
from Ex05_basf2_simulation import SimulationTask


# %%
# Here, I make use of the ``requires`` decorator to require the ``SimulationTask``.
# This functionality is very useful when the dependencies of tasks are very
# easily defined. The task with the ``requires`` decorator will automatically
# create the ``requires`` method that returns the required task. Additionally,
# the task also inherits all the parameters of the required task.


# %%
@b2luigi.requires(SimulationTask)
class ReconstructionTask(Basf2PathTask):
    result_dir = "results/Reconstruction"
    batch_system = "local"

    def output(self):
        yield self.add_to_output("reconstruction.root")

    def create_path(self):
        main = b2.Path()

        # Read the simulated events that were created in the simulation task.
        main.add_module("RootInput", inputFileNames=self.get_input_file_names("simulation.root"))

        # Run reconstruction
        main.add_module("Gearbox")
        main.add_module("Geometry")
        re.add_reconstruction(main)

        # Add MDST output.
        mdst.add_mdst_output(path=main, filename=self.get_output_file_name("reconstruction.root"))
        return main


# %%
# As already pointed out in the second example, here we only need to
# call the ``ReconstructionTask``: the scheduler will resolves all the
# required dependencies.
if __name__ == "__main__":
    b2luigi.process(ReconstructionTask(n_events=10))



import os

from ..helpers import B2LuigiTestCase

import b2luigi


class TemporaryWrapperTestCase(B2LuigiTestCase):
    def test_temporary_file_name(self):
        base_name = "a-complex_^name&114%"

        targetA = b2luigi.LocalTarget(f"{base_name}.foo")
        tmp_nameA = targetA.tmp_name
        self.assertTrue(tmp_nameA.startswith(f"{base_name}-luigi-tmp-"))
        self.assertTrue(tmp_nameA.endswith(".foo"))
        self.assertFalse(tmp_nameA.endswith("..foo"))
        self.assertFalse(tmp_nameA.endswith("foo."))

        targetB = b2luigi.LocalTarget(f"{base_name}.foo.bar")
        tmp_nameB = targetB.tmp_name
        self.assertTrue(tmp_nameB.startswith(f"{base_name}-luigi-tmp-"))
        self.assertTrue(tmp_nameB.endswith(".foo.bar"))
        self.assertFalse(tmp_nameB.endswith("..foo.bar"))
        self.assertFalse(tmp_nameB.endswith(".foo..bar"))

        targetC = b2luigi.LocalTarget(f"{base_name}")
        tmp_nameC = targetC.tmp_name
        self.assertTrue(tmp_nameC.startswith(f"{base_name}-luigi-tmp-"))
        self.assertFalse(tmp_nameC.endswith("."))

    def test_failed_temporary_files(self):
        self.call_file("core/temporary_task_1.py")

        self.assertFalse(os.path.exists("test.txt"))

    def test_good_temporary_files(self):
        self.call_file("core/temporary_task_2.py")

        self.assertTrue(os.path.exists("test.txt"))

        with open("test.txt", "r") as f:
            self.assertEqual(f.read(), "Test")

    def test_reset_output(self):
        class MyTask(b2luigi.Task):
            def output(self):
                yield self.add_to_output("test.txt")

            @b2luigi.on_temporary_files
            def run(self):
                with open(self.get_output_file_name("test.txt"), "w") as f:
                    f.write("Test")

        task = MyTask()

        self.assertTrue(task.get_output_file_name("test.txt").startswith(self.test_dir))

    def test_reset_input(self):
        class TaskA(b2luigi.Task):
            parameter_a = b2luigi.IntParameter()

            def output(self):
                yield self.add_to_output("resA.txt")

            @b2luigi.on_temporary_files
            def run(self):
                with open(self.get_output_file_name("resA.txt"), "w") as f:
                    f.write(f"TaskA: {self.parameter_a}")

        class MyTask(b2luigi.Task):
            def requires(self):
                return TaskA(parameter_a=1)

        task = MyTask()
        self.assertTrue(task.get_input_file_name("resA.txt").startswith(self.test_dir))



"""
.. _exercise08_label:

Scaling up your analysis
========================

.. hint::
    This example demonstrates how to locally scale the analysis task from
    the previous example :ref:`exercise07_label`. This example is very similar
    to the basic ``luigi`` wrapper task example in :ref:`exercise03_label`. Additionally, we
    introduce the ``inherits`` method to inherit parameters from another task
    without requiring the task.
"""

import b2luigi
from Ex07_basf2_analysis import AnalysisTask


# %%
# The ``inherits`` decorater works similar to the already known ``requires``
# decorator. It will also propagate the parameters of the inherited task to
# the inheriting task. The difference is that the inheriting task does not
# require the inherited task. It can e.g. be used in tasks that merge the
# output of the tasks they require. These merger tasks don't need the
# parameter they resolve anymore but should keep the same order of parameters,
# therefore simplifying the directory


# %%
@b2luigi.inherits(AnalysisTask, without="identifier")
class AnalysisWrapperTask(b2luigi.WrapperTask):
    n_loops = b2luigi.IntParameter(default=1)

    def requires(self):
        for i in range(self.n_loops):
            yield AnalysisTask(n_events=self.n_events, identifier=f"loop_{i}")


if __name__ == "__main__":
    b2luigi.process(AnalysisWrapperTask(n_events=10, n_loops=5), workers=5)



import unittest
from unittest.mock import Mock
from b2luigi.batch.workers import SendJobWorker, BatchSystems


class TestSendJobWorker(unittest.TestCase):
    def setUp(self):
        self.worker = SendJobWorker()

    def test_create_task_process_lsf(self):
        task = Mock()
        task.batch_system = "lsf"
        batch_system = self.worker.detect_batch_system(task)
        self.assertEqual(batch_system, BatchSystems.lsf)

    def test_create_task_process_htcondor(self):
        task = Mock()
        task.batch_system = "htcondor"
        batch_system = self.worker.detect_batch_system(task)
        self.assertEqual(batch_system, BatchSystems.htcondor)

    def test_create_task_process_slurm(self):
        task = Mock()
        task.batch_system = "slurm"
        batch_system = self.worker.detect_batch_system(task)
        self.assertEqual(batch_system, BatchSystems.slurm)

    def test_create_task_process_gbasf2(self):
        task = Mock()
        task.batch_system = "gbasf2"
        batch_system = self.worker.detect_batch_system(task)
        self.assertEqual(batch_system, BatchSystems.gbasf2)

    def test_create_task_process_test(self):
        task = Mock()
        task.batch_system = "test"
        batch_system = self.worker.detect_batch_system(task)
        self.assertEqual(batch_system, BatchSystems.test)

    def test_create_process_auto(self):
        task = Mock()
        task.batch_system = "auto"
        batch_system = self.worker.detect_batch_system(task)
        self.assertIn(batch_system, [BatchSystems.lsf, BatchSystems.htcondor, BatchSystems.slurm, BatchSystems.local])

    def test_create_task_process_not_implemented(self):
        task = Mock()
        task.batch_system = "unknown"
        with self.assertRaises(ValueError):
            self.worker._create_task_process(task)


if __name__ == "__main__":
    unittest.main()



import unittest
import subprocess
from unittest.mock import patch, MagicMock
from b2luigi.batch.processes.apptainer import ApptainerProcess
from b2luigi.batch.processes import JobStatus
from .batch_task_1 import MyTask


class MyApptainerTask(MyTask):
    apptainer_image = "/cvmfs/belle.cern.ch/images/belle2-base-el9"
    apptainer_mounts = ["/cvmfs"]
    apptainer_mount_defaults = True
    apptainer_additional_params = "--cleanenv"


class TestApptainerProcess(unittest.TestCase):
    def setUp(self):
        self.mock_task = MyApptainerTask("some_parameter")
        self.mock_scheduler = MagicMock()
        self.mock_result_queue = MagicMock()
        self.mock_worker_timeout = MagicMock()
        self.process = ApptainerProcess(
            task=self.mock_task,
            scheduler=self.mock_scheduler,
            result_queue=self.mock_result_queue,
            worker_timeout=self.mock_worker_timeout,
        )

    @patch("b2luigi.batch.processes.apptainer.create_cmd_from_task")
    @patch("b2luigi.batch.processes.apptainer.create_apptainer_command")
    @patch("subprocess.Popen")
    def test_start_job(self, mock_popen, mock_create_apptainer_command, mock_create_cmd_from_task):
        mock_create_cmd_from_task.return_value = ["echo", "hello"]
        mock_create_apptainer_command.return_value = "apptainer exec echo hello"
        mock_process = MagicMock()
        mock_popen.return_value = mock_process

        self.process.start_job()

        mock_create_cmd_from_task.assert_called_once_with(self.mock_task)
        mock_create_apptainer_command.assert_called_once_with("echo hello", task=self.mock_task)
        mock_popen.assert_called_once_with("apptainer exec echo hello", stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.assertEqual(self.process._process, mock_process)

    @patch("b2luigi.batch.processes.apptainer.get_log_file_dir")
    def test_write_output(self, mock_get_log_file_dir):
        mock_get_log_file_dir.return_value = "/tmp"
        self.process._stdout = b"stdout content"
        self.process._stderr = b"stderr content"

        self.process._write_output()

        with open("/tmp/stdout", "r") as f:
            self.assertEqual(f.read(), "stdout content")

        with open("/tmp/stderr", "r") as f:
            self.assertEqual(f.read(), "stderr content")

    def test_get_job_status_running(self):
        self.process._process = MagicMock()
        self.process._process.poll.return_value = None

        status = self.process.get_job_status()
        self.assertEqual(status, JobStatus.running)

    @patch("b2luigi.batch.processes.apptainer.ApptainerProcess._write_output")
    def test_get_job_status_successful(self, mock_write_output):
        mock_write_output.return_value = True
        self.process._process = MagicMock()
        self.process._process.poll.return_value = 0
        self.process._process.communicate.return_value = (b"stdout content", b"stderr content")
        self.process._process.returncode = 0

        status = self.process.get_job_status()
        self.assertEqual(status, JobStatus.successful)

    @patch("b2luigi.batch.processes.apptainer.ApptainerProcess._write_output")
    def test_get_job_status_aborted(self, mock_write_output):
        self.process._process = None
        status = self.process.get_job_status()
        self.assertEqual(status, JobStatus.aborted)

        mock_write_output.return_value = True
        self.process._process = MagicMock()
        self.process._process.poll.return_value = 1
        self.process._process.communicate.return_value = (b"stdout content", b"stderr content")
        self.process._process.returncode = 1

        status = self.process.get_job_status()
        self.assertEqual(status, JobStatus.aborted)

    def test_terminate_job(self):
        self.process._process = MagicMock()
        self.process.terminate_job()
        self.process._process.terminate.assert_called_once()



import pickle

from basf2 import conditions as b2conditions, logging
from basf2.pickle_path import serialize_path
from variables import variables as vm


def get_alias_dict_from_variable_manager():
    """
    Extracts a dictionary of alias names and their corresponding variable names from the
    internal state of the variable manager.

    This function iterates over all alias names managed by the variable manager, retrieves
    the associated variable for each alias, and constructs a dictionary where the keys
    are the alias names (as strings) and the values are the names of the corresponding
    variables.

    Returns:
        dict: A dictionary where keys are alias names (str) and values are the names of
        the corresponding variables (str).
    """
    alias_dictionary = {str(alias_name): vm.getVariable(str(alias_name)).name for alias_name in vm.getAliasNames()}
    return alias_dictionary


def write_path_and_state_to_file(basf2_path, file_path):
    """
    Serialize a ``basf2`` path, variable aliases, and global tags to a file.

    This function is a variant of ``basf2.pickle_path.write_path_to_file`` with
    additional serialization of ``basf2`` variable aliases and global tags. The aliases
    are extracted from the current state of the variable manager singleton, and the
    global tags are taken from the `b2conditions` module. Both must be set in the
    python/``basf2`` process before calling this function.

    Args:
        basf2_path: The ``basf2`` path object to serialize.
        file_path: The file path where the serialized pickle object will be written.
    """
    with open(file_path, "bw") as pickle_file:
        serialized = serialize_path(basf2_path)
        serialized["aliases"] = get_alias_dict_from_variable_manager()
        serialized["globaltags"] = b2conditions.globaltags
        serialized["log_level"] = logging.log_level
        # serialized["conditions"] = b2conditions
        pickle.dump(serialized, pickle_file)



import subprocess
import pathlib
import re
import getpass
import enum
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_exponential

from luigi.parameter import _no_value
from b2luigi.core.settings import get_setting
from b2luigi.batch.processes import BatchProcess, JobStatus
from b2luigi.batch.cache import BatchJobStatusCache
from b2luigi.core.utils import get_log_file_dir, get_task_file_dir
from b2luigi.core.executable import create_executable_wrapper


class SlurmJobStatusCache(BatchJobStatusCache):
    sacct_disabled = None

    @retry(
        retry=retry_if_exception_type(subprocess.CalledProcessError),
        stop=stop_after_attempt(4),
        wait=wait_exponential(multiplier=2, min=2, exp_base=3),  # 2, 6, 18 seconds
        reraise=True,
    )
    def _ask_for_job_status(self, job_id: int = None):
        """
        With Slurm, you can check the progress of your jobs using the ``squeue`` command.
        If no ``jobID`` is given as argument, this command shows you the status of all queued jobs.

        Sometimes it might happen that a job is completed in between the status checks. Then its final status
        can be found using ``sacct`` (works mostly in the same way as ``squeue``).

        If in the unlikely case the server has the Slurm accounting disabled, then the `scontrol` command is used as a last
        resort to access the jobs history. This is the fail safe command as the scontrol by design only holds onto a jobs
        information for a short period of time after completion. The time between status checks is sufficiently short however
        so the scontrol command should still have the jobs information on hand.

        All three commands are used in order to find out the :obj:`SlurmJobStatus`.
        """
        # https://slurm.schedmd.com/squeue.html
        user = getpass.getuser()
        q_cmd = ["squeue", "--noheader", "--user", user, "--format", "'%i %T'"] + (
            ["--job", str(job_id)] if job_id else []
        )
        output = subprocess.check_output(q_cmd)

        output = output.decode()
        seen_ids = self._fill_from_output(output)
        # If no job_id was passed, then exit
        if not job_id:
            return
        # If the specified job can not be found in the squeue output, we need to request its history from the slurm job accounting log
        # We also check that the working Slurm server has the Slurm accounting storage active
        if job_id not in seen_ids and not self._check_if_sacct_is_disabled_on_server():
            # https://slurm.schedmd.com/sacct.html
            history_cmd = [
                "sacct",
                "--noheader",
                "-X",
                "--user",
                user,
                "--format=JobID,State",
                "--job",
                str(job_id),
            ]
            output = subprocess.check_output(history_cmd)
            output = output.decode()
            self._fill_from_output(output)

        # If the Slurm accounting storage is disabled, we resort to the scontrol command
        elif job_id not in seen_ids and self._check_if_sacct_is_disabled_on_server():
            output = subprocess.check_output(["scontrol", "show", "job", str(job_id)])
            output = output.decode()

            # Extract the job state from the output
            re_output = re.search(r"JobState=([A-Z_]+)", output)
            if re_output:
                state_string = re_output.group(1)
                self[job_id] = self._get_SlurmJobStatus_from_string(state_string)

            # the specified job cannot be found on the slurm system. Return a failed.
        else:
            self[job_id] = SlurmJobStatus.failed

    def _fill_from_output(self, output: str) -> set:
        """
        Parses the output of a Slurm command to extract job IDs and their states,
        updating the internal job status mapping and returning a set of seen job IDs.

        Args:
            output (str): The output string from a Slurm command, expected to be
                          formatted as '<job id> <state>' per line.

        Returns:
            set: A set of job IDs that were parsed from the output.

        Raises:
            AssertionError: If a line in the output does not contain exactly two
                            entries (job ID and state).

        Notes:
            - If the output is empty, an empty set is returned.
            - Lines in the output that are empty or contain unexpected formatting
              are skipped.
            - Job states with a '+' suffix (e.g., 'CANCELLED+') are normalized by
              stripping the '+' character.
        """
        seen_ids = set()

        # If the output is empty return an empty set
        if not output:
            return seen_ids

        # If no jobs exist then output=='' and this loop does not see any id's
        for job_info_str in output.split("\n"):
            if not job_info_str:
                continue  # When splitting by \n, the final entry of the list is likely an empty string
            job_info = job_info_str.strip("'").split()

            # We have formatted the squeue and sacct outputs to be '<job id> <state>'
            # hence we expect there to always be two entries in the list
            assert (
                len(job_info) == 2
            ), "Unexpected behaviour has occurred whilst retrieving job information. There may be an issue with the sqeue, sacct or scontrol commands."
            id, state_string = job_info
            id = int(id)
            # Manually cancelling jobs gives the state 'CANCELLED+'
            state_string = state_string.strip("+")
            self[id] = self._get_SlurmJobStatus_from_string(state_string)
            seen_ids.add(id)

        return seen_ids

    def _get_SlurmJobStatus_from_string(self, state_string: str) -> str:
        """
        Converts a state string into a :obj:`SlurmJobStatus` enumeration value.

        Args:
            state_string (str): The state string to be converted.

        Returns:
            str: The corresponding :obj:`SlurmJobStatus` value.

        Raises:
            KeyError: If the provided state string does not match any valid :obj:`SlurmJobStatus`.
        """
        try:
            state = SlurmJobStatus(state_string)
        except KeyError:
            raise KeyError(f"The state {state_string} could not be found in the SlurmJobStatus states")
        return state

    def _check_if_sacct_is_disabled_on_server(self) -> bool:
        """
        Checks if the Slurm accounting command ``sacct`` is disabled on the system.

        This method determines whether the ``sacct`` command is unavailable or
        disabled by attempting to execute it and analyzing the output. The result
        is cached in the ``self.sacct_disabled`` attribute to avoid repeated checks.

        Returns:
            bool: True if ``sacct`` is disabled on the system, ``False`` otherwise.
        """
        if self.sacct_disabled is None:
            # Don't continually call the function, instead call it once and set self.sacct_disabled
            output = subprocess.run(["sacct"], capture_output=True)

            # If the call to 'sacct' returns an error code 1 and checking the stderr returns 'Slurm accounting storage is disabled'
            self.sacct_disabled = (
                output.returncode == 1 and output.stderr.strip().decode() == "Slurm accounting storage is disabled"
            )
        return self.sacct_disabled


class SlurmJobStatus(enum.Enum):
    """
    See https://slurm.schedmd.com/job_state_codes.html
    TODO: make this a StrEnum with python>=3.11

    Attributes:
        completed (str): The job has completed successfully.
        pending (str): The job is waiting to be scheduled.
        running (str): The job is currently running.
        suspended (str): The job has been suspended.
        preempted (str): The job has been preempted by another job.
        completing (str): The job is in the process of completing.

        boot_fail (str): The job failed during the boot process.
        cancelled (str): The job was cancelled by the user or system.
        deadline (str): The job missed its deadline.
        node_fail (str): The job failed due to a node failure.
        out_of_memory (str): The job ran out of memory.
        failed (str): The job failed for an unspecified reason.
        timeout (str): The job exceeded its time limit.
    """

    # successful:
    completed = "COMPLETED"

    # running:
    pending = "PENDING"
    running = "RUNNING"
    suspended = "SUSPENDED"
    preempted = "PREEMPTED"
    completing = "COMPLETING"

    # failed:
    boot_fail = "BOOT_FAIL"
    cancelled = "CANCELLED"
    deadline = "DEADLINE"
    node_fail = "NODE_FAIL"
    out_of_memory = "OUT_OF_MEMORY"
    failed = "FAILED"
    timeout = "TIMEOUT"

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, SlurmJobStatus):
            return self.value == other.value
        raise TypeError(
            "The equivalence of a SlurmJobStatus can only be checked with a string or another SlurmJobStatus object."
        )


_batch_job_status_cache = SlurmJobStatusCache()


class SlurmProcess(BatchProcess):
    """
    Reference implementation of the batch process for a Slurm batch system.

    Additional to the basic batch setup (see :ref:`batch-label`), additional
    Slurm-specific things are:

    * Please note that most of the Slurm batch farms by default copy the user environment from
      the submission node to the worker machine. As this can lead to different results when running
      the same tasks depending on your active environment, you probably want to pass the argument
      ``export=NONE``. This ensures that a reproducible environment is used. You can provide an
      ``env_script``, an ``env`` :meth:`setting <b2luigi.set_setting>`, and/or a different
      ``executable`` to create the environment necessary for your task.

    * Via the ``slurm_settings`` setting you can provide a dict for additional options, such as
      requested memory etc. Its value has to be a dictionary
      containing Slurm settings as key/value pairs. These options will be written into the job
      submission file. For an overview of possible settings refer to the `Slurm documentation
      <https://slurm.schedmd.com/sbatch.html#>_` and the documentation of the cluster you are using.

    * Same as for the :ref:`lsf` and :ref:`htcondor`, the ``job_name`` setting allows giving a meaningful
      name to a group of jobs. If you want to be task-instance-specific, you can provide the ``job-name``
      as an entry in the ``slurm_settings`` dict, which will override the global ``job_name`` setting.
      This is useful for manually checking the status of specific jobs with

      .. code-block:: bash

        squeue --name <job name>

    Example:

        .. literalinclude:: ../../examples/slurm/slurm_example.py
           :linenos:
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._batch_job_id = None

    def get_job_status(self):
        """
        Determine the status of a batch job based on its Slurm job status.

        Returns:
            JobStatus: The status of the job, which can be one of the following:
                - :meth:`JobStatus.successful <b2luigi.process.JobStatus.successful>`: If the job has completed successfully.
                - :meth:`JobStatus.running <b2luigi.process.JobStatus.running>`: If the job is currently running, pending, suspended, preempted, or completing.
                - :meth:`JobStatus.aborted <b2luigi.process.JobStatus.aborted>`: If the job has failed, been cancelled, exceeded its deadline, encountered a node failure,
                  ran out of memory, timed out, or if the job ID is not found.

        Raises:
            ValueError: If the Slurm job status is unknown or not handled.
        """
        if not self._batch_job_id:
            return JobStatus.aborted
        try:
            job_status = _batch_job_status_cache[self._batch_job_id]
        except KeyError:
            return JobStatus.aborted

        # See https://slurm.schedmd.com/job_state_codes.html
        if job_status in [SlurmJobStatus.completed]:
            return JobStatus.successful
        if job_status in [
            SlurmJobStatus.pending,
            SlurmJobStatus.running,
            SlurmJobStatus.suspended,
            SlurmJobStatus.preempted,
            SlurmJobStatus.completing,
        ]:
            return JobStatus.running
        if job_status in [
            SlurmJobStatus.boot_fail,
            SlurmJobStatus.cancelled,
            SlurmJobStatus.deadline,
            SlurmJobStatus.node_fail,
            SlurmJobStatus.out_of_memory,
            SlurmJobStatus.failed,
            SlurmJobStatus.timeout,
        ]:
            return JobStatus.aborted
        raise ValueError(f"Unknown Slurm Job status: {job_status}")

    def start_job(self):
        """
        Starts a job by submitting the Slurm submission script.

        This method creates a Slurm submit file and submits it using the ``sbatch`` command.
        After submission, it parses the output to extract the batch job ID.

        Raises:
            RuntimeError: If the batch submission fails or the job ID cannot be extracted.

        Attributes:
            self._batch_job_id (int): The ID of the submitted Slurm batch job.
        """
        submit_file = self._create_slurm_submit_file()

        # Slurm submit needs to be called in the folder of the submit file
        path = pathlib.Path(submit_file)
        output = subprocess.check_output(["sbatch", path.name], cwd=path.parent)

        output = output.decode()
        match = re.search(r"[0-9]+", output)
        if not match:
            raise RuntimeError("Batch submission failed with output " + output)
        self._batch_job_id = int(match.group(0))

    def terminate_job(self):
        """
        Terminates a batch job if a job ID is available.

        This method checks if a batch job ID is set. If it is, it executes the
        ``scancel`` command to terminate the job associated with the given batch
        job ID. The command's output is suppressed.
        """
        if not self._batch_job_id:
            return
        subprocess.run(["scancel", str(self._batch_job_id)], stdout=subprocess.DEVNULL)

    def _create_slurm_submit_file(self):
        """
        Creates a Slurm submit file for the current task.

        This method generates a Slurm batch script that specifies the necessary
        configurations for submitting a job to a Slurm workload manager.

        Returns:
            pathlib.Path: The path to the generated Slurm submit file.

        Note:
            - The ``stdout`` and ``stderr`` log files are created in the task's log directory. See :obj:`get_log_file_dir`.
            - The Slurm settings are specified in the ``slurm_settings`` setting, which is a dictionary of key-value pairs.
            - The ``job_name`` setting can be used to specify a meaningful name for the job.
            - The executable is created with :obj:`create_executable_wrapper`.
            - The submit file is named `slurm_parameters.sh` and is created in the task's output directory (:obj:`get_task_file_dir`).
        """
        submit_file_content = ["#!/usr/bin/bash"]

        # Specify where to write the log to
        log_file_dir = pathlib.Path(get_log_file_dir(self.task))
        log_file_dir.mkdir(parents=True, exist_ok=True)

        stdout_log_file = (log_file_dir / "stdout").resolve()
        submit_file_content.append(f"#SBATCH --output={stdout_log_file}")

        stderr_log_file = (log_file_dir / "stderr").resolve()
        submit_file_content.append(f"#SBATCH --error={stderr_log_file}")

        # Specify additional settings
        # A default value of None requires that the user must set the setting. We therefore use luigi.parameters._no_value.
        general_settings = get_setting("slurm_settings", default=_no_value)
        if general_settings == _no_value:
            general_settings = {}

        task_slurm_settings = get_setting("slurm_settings", task=self.task, default=_no_value)
        if task_slurm_settings != _no_value:
            general_settings.update(task_slurm_settings)

        job_name = get_setting("job_name", task=self.task, default=False)
        if job_name is not False:
            general_settings.setdefault("job-name", job_name)

        for key, item in general_settings.items():
            submit_file_content.append(f"#SBATCH --{key}={item}")

        # Specify the executable
        executable_file = create_executable_wrapper(self.task)
        submit_file_content.append(f"exec {pathlib.Path(executable_file).resolve()}")

        # Now we can write the submit file
        output_path = pathlib.Path(get_task_file_dir(self.task))
        submit_file_path = output_path / "slurm_parameters.sh"

        output_path.mkdir(parents=True, exist_ok=True)
        with open(submit_file_path, "w") as submit_file:
            submit_file.write("\n".join(submit_file_content))
        return submit_file_path



"""
Test helper functions for :py:class:`HTCondorProcess`.
"""

import json
import os
import subprocess
import unittest
from unittest import mock

import b2luigi
from b2luigi.batch.processes.htcondor import HTCondorJobStatusCache, HTCondorProcess

from ..helpers import B2LuigiTestCase
from .batch_task_1 import MyTask


class TestHTCondorCreateSubmitFile(B2LuigiTestCase):
    def _get_htcondor_submit_file_string(self, task):
        # the _create_htcondor_submit_file is a method of the ``HTCondorProcess`` class, but it only uses its
        # class to obtain ``self.task``, to it's sufficient to provide a mock class for ``self``, `
        htcondor_mock_process = mock.Mock()
        task.get_task_file_dir = lambda: self.test_dir
        task.get_log_file_dir = lambda: self.test_dir
        htcondor_mock_process.task = task
        #  create submit file
        HTCondorProcess._create_htcondor_submit_file(htcondor_mock_process)
        # read submit file and return string
        submit_file_path = os.path.join(self.test_dir, "job.submit")
        with open(submit_file_path, "r") as submit_file:
            return submit_file.read()

    def test_minimal_submit_file(self):
        """
        Minimal  submit file should have expected shape:

            output = ...
            error = ..
            log = ..
            executable = executable_wrapper.sh
            queue 1
        """
        submit_file_lines = self._get_htcondor_submit_file_string(MyTask("some_parameter")).splitlines()
        self.assertIn("output = ", submit_file_lines[0])
        self.assertIn("error = ", submit_file_lines[1])
        self.assertIn("log = ", submit_file_lines[2])
        self.assertEqual("executable = executable_wrapper.sh", submit_file_lines[3])
        self.assertEqual("queue 1", submit_file_lines[4])

    def test_not_setting_job_name(self):
        submit_file_string = self._get_htcondor_submit_file_string(MyTask("some_parameter"))
        self.assertNotIn("JobBatchName", submit_file_string)

    def test_set_job_name_via_task_attribute(self):
        task = MyTask("some_parameter")
        task.job_name = "some_job_name"
        submit_file_lines = self._get_htcondor_submit_file_string(task).splitlines()
        self.assertIn("JobBatchName = some_job_name", submit_file_lines)

        b2luigi.set_setting("job_name", "some_job_name")
        submit_file_lines = self._get_htcondor_submit_file_string(MyTask("some_parameter")).splitlines()
        b2luigi.clear_setting("job_name")
        self.assertIn("JobBatchName = some_job_name", submit_file_lines)

    def test_set_job_name_is_overriden_by_htcondor_settings(self):
        """
        ``job_name`` is a global setting, but if the ``JobBatchName`` is set explicitly via the settings, we
        want that to override the global setting
        """
        task = MyTask("some_parameter")
        task.job_name = "job_name_global"
        htcondor_settings = {"JobBatchName": "job_name_htcondor"}
        b2luigi.set_setting("htcondor_settings", htcondor_settings)
        submit_file_lines = self._get_htcondor_submit_file_string(task).splitlines()
        b2luigi.clear_setting("htcondor_settings")
        self.assertNotIn("JobBatchName = job_name_global", submit_file_lines)
        self.assertIn("JobBatchName = job_name_htcondor", submit_file_lines)


class TestHTCondorJobStatusCache(unittest.TestCase):
    def setUp(self):
        mock_status_dicts = [
            {
                "JobStatus": 4,  # completed
                "ExitCode": 0,  # success
                "ClusterId": 42,
            },
            {
                "JobStatus": 2,  # running
                "ClusterId": 43,
            },
            {
                "JobStatus": 999,  # failed
                "ClusterId": 44,
            },
        ]
        self.mock_status_json = json.dumps(mock_status_dicts).encode()

        self.htcondor_job_status_cache = HTCondorJobStatusCache()

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_does_2_retries(self, mock_check_output):
        """Test the ``_ask_for_job_status`` recovers after two condor_q failures."""

        # make check_output fail 2 times before  return status dict
        n_fail = 2
        mock_check_output.side_effect = n_fail * [subprocess.CalledProcessError(1, ["mock", "command"])] + [
            self.mock_status_json
        ]
        self.htcondor_job_status_cache._ask_for_job_status()
        self.assertEqual(mock_check_output.call_count, 3)

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_no_retries_on_success(self, mock_check_output):
        """Test the ``_ask_for_job_status`` is only called once (no retries) when everything works."""

        # make check_output fail 2 times before  return status dict
        mock_check_output.return_value = self.mock_status_json
        self.htcondor_job_status_cache._ask_for_job_status()
        self.assertEqual(mock_check_output.call_count, 1)

    @mock.patch("subprocess.check_output")
    def test_ask_for_job_status_fails_after_4_condor_q_failures(self, mock_check_output):
        """Test the ``_ask_for_job_status`` does not do more than 3 retries"""

        # make check_output fail 2 times before  return status dict
        n_fail = 4
        mock_check_output.side_effect = n_fail * [subprocess.CalledProcessError(1, ["mock", "command"])] + [
            self.mock_status_json
        ]
        with self.assertRaises(subprocess.CalledProcessError):
            self.htcondor_job_status_cache._ask_for_job_status()



from contextlib import ExitStack
from functools import wraps
from multiprocessing.pool import ThreadPool
from typing import Optional

from b2luigi import Task
from b2luigi.core import utils
from b2luigi.core.settings import get_setting


class TemporaryFileContextManager(ExitStack):
    """
    A context manager for managing temporary file paths for a given task. This class overrides
    the task's methods for retrieving input and output file names, ensuring that temporary file
    paths are used during the context's lifetime. Upon exiting the context, the original methods
    are restored.

    Usage:
        This class is not meant to be used directly. Instead, it is used within the :obj:`on_temporary_files`
    """

    def __init__(self, task: Task):
        super().__init__()

        self._task = task
        self._task_output_function = task.get_output_file_name
        self._task_input_function = task.get_input_file_names
        self._task_all_input_function = task.get_all_input_file_names
        self._task_input_function_from_dict = task.get_input_file_names_from_dict

        self._open_output_files = {}
        self._open_input_files = {}

    def __enter__(self):
        """
        This method is called when entering the context manager. It overrides the task's methods
        for retrieving input and output file names, replacing them with custom implementations
        that handle temporary file paths.

        **Redefinitions**:

        - ``get_output_file_name``

          Retrieves or creates a temporary output file for the given key.

          If the file corresponding to the specified key is not already open, this method
          generates a temporary file path using the task's :meth:`b2luigi.Task._get_output_target` and
          :meth:`b2luigi.Target.temporary_path` methods, and opens the file within the context of the
          task.

          Args:
              key (str): The unique identifier for the output file.

              tmp_file_kwargs: Additional keyword arguments to be passed to the :meth:`b2luigi.Target.temporary_path` method.

          Returns:
              The opened temporary file associated with the given key.

        - ``get_input_file_names``

          Retrieves the input file names associated with a given key, ensuring that the files are
          temporarily available for processing. Behaves the same as :meth:`b2luigi.Task.get_input_file_names`.

          Args:
              key (str): The identifier for the input files to retrieve.

              tmp_file_kwargs: Additional keyword arguments to pass to the :meth:`b2luigi.Target.get_temporary_input` method of the target.

          Returns:
              list: A list of opened temporary input files corresponding to the given key.

          Notes:
              - If the ``n_download_threads`` setting is specified, the input files are fetched
                concurrently using a thread pool.
              - If ``n_download_threads`` is not specified, the input files are fetched sequentially.

        - ``get_all_input_file_names``
            Retrieves all input file names from the task's input, ensuring that all files are temporarily available for processing.
            Behaves the same as :meth:`b2luigi.Task.get_all_input_file_names`.

            Args:
                tmp_file_kwargs: Additional keyword arguments to pass to the :meth:`b2luigi.Target.get_temporary_input` method of the target.

            Returns:
                list: A list of all opened temporary input files from the task's input.

            Notes:
              - If the ``n_download_threads`` setting is specified, the input files are fetched concurrently
                using a thread pool.
              - If ``n_download_threads`` is not specified, the input files are fetched sequentially.

        - ``get_input_file_names_from_dict``

          Retrieves input file names from a dictionary structure, handling temporary file paths.
          Behaves the same as :meth:`b2luigi.Task.get_input_file_names_from_dict`.

          Args:
              requirement_key (str): The key used to identify the required input in the task's input dictionary.

              key (Optional[str], optional): A specific key to extract targets from the target dictionary. If ``None``, all targets are retrieved. Defaults to ``None``.

              tmp_file_kwargs: Additional keyword arguments passed to the :meth:`b2luigi.Target.get_temporary_input` method for generating temporary file paths.

          Returns:
              list: A list of temporary file paths corresponding to the input files.

          Raises:
              KeyError: If the specified `key` is not found in the target dictionary.
        """

        def get_output_file_name(key: str, **tmp_file_kwargs):
            if key not in self._open_output_files:
                target = self._task._get_output_target(key)
                temporary_path = target.temporary_path(task=self._task, **tmp_file_kwargs)
                self._open_output_files[key] = self.enter_context(temporary_path)

            return self._open_output_files[key]

        self._task.get_output_file_name = get_output_file_name

        def get_input_file_names(key: str, **tmp_file_kwargs):
            """ """
            if key not in self._open_input_files:
                targets = self._task._get_input_targets(key)
                self._open_input_files[key] = []
                n_download_threads = get_setting("n_download_threads", default=2, task=self._task)
                if n_download_threads is not None:
                    with ThreadPool(n_download_threads) as pool:
                        self._open_input_files[key] = pool.map(
                            lambda target: self.enter_context(
                                target.get_temporary_input(task=self._task, **tmp_file_kwargs)
                            ),
                            targets,
                        )
                else:
                    for target in targets:  # TODO: This does not work in wrapping tasks
                        temporary_path = target.get_temporary_input(task=self._task, **tmp_file_kwargs)
                        self._open_input_files[key].append(self.enter_context(temporary_path))

            return self._open_input_files[key]

        self._task.get_input_file_names = get_input_file_names

        def get_all_input_file_names(**tmp_file_kwargs):
            input_dict = utils.flatten_to_dict_of_lists(self._task.input())
            n_download_threads = get_setting("n_download_threads", default=2, task=self._task)
            for key, value in input_dict.items():
                if key not in self._open_input_files:
                    self._open_input_files[key] = []
                    if n_download_threads is not None:
                        with ThreadPool(n_download_threads) as pool:
                            self._open_input_files[key] = pool.map(
                                lambda target: self.enter_context(
                                    target.get_temporary_input(task=self._task, **tmp_file_kwargs)
                                ),
                                value,
                            )
                    else:
                        for target in value:
                            temporary_path = target.get_temporary_input(task=self._task, **tmp_file_kwargs)
                            self._open_input_files[key].append(self.enter_context(temporary_path))

            return list(self._open_input_files.values())

        self._task.get_all_input_file_names = get_all_input_file_names

        def get_input_file_names_from_dict(requirement_key: str, key: Optional[str] = None, **tmp_file_kwargs):
            internal_key = f"{requirement_key}_{str(key)}"
            if internal_key not in self._open_input_files:
                # Expected output of task.input is {requirement_key: [generators, ...]}
                target_generator_dict = utils.flatten_to_dict_of_lists(self._task.input()[requirement_key])
                # Now, the output is structured as {key: [target1, target2, ...]}

                self._open_input_files[internal_key] = []

                if key is not None and key in target_generator_dict.keys():
                    targets = target_generator_dict[key]

                # If key is None, we want to get all targets from the dictionary
                else:
                    targets = [item for sublist in target_generator_dict.values() for item in sublist]

                for target in targets:
                    temporary_path = target.get_temporary_input(task=self._task, **tmp_file_kwargs)
                    self._open_input_files[internal_key].append(self.enter_context(temporary_path))

            return self._open_input_files[internal_key]

        self._task.get_input_file_names_from_dict = get_input_file_names_from_dict

    def __exit__(self, *exc_details):
        super().__exit__(*exc_details)

        self._task.get_output_file_name = self._task_output_function
        self._task.get_input_file_names = self._task_input_function
        self._task.get_all_input_file_names = self._task_all_input_function
        self._task.get_input_file_names_from_dict = self._task_input_function_from_dict


def on_temporary_files(run_function):
    """
    Wrapper for decorating a task's run function to use temporary files as outputs.

    A common problem when using long running tasks in ``luigi`` is the so called thanksgiving bug
    (see https://www.arashrouhani.com/luigi-budapest-bi-oct-2015/#/21).
    It occurs, when you define an output of a task and in its run function,
    you create this output before filling it with content
    (maybe even only after a long lasting calculation).
    It may happen, that during the creation of the output and the finish of the calculation
    some other tasks checks if the output is already there, finds it and assumes,
    that the task is already finished (although there is probably only non-sense in the file
    so far).

    A solution is already given by luigi itself, when using the ``temporary_path()`` function
    of the file system targets, which is really nice!
    Unfortunately, this means you have to open all your output files with a context manager
    and this is very hard to do if you have external tasks also (because they will
    probably use the output file directly instead of the temporary file version of if).

    This wrapper simplifies the usage of the temporary files:

    .. code-block:: python

      import b2luigi

      class MyTask(b2luigi.Task):
          def output(self):
              yield self.add_to_output("test.txt")

          @b2luigi.on_temporary_files
          def run(self):
              with open(self.get_output_file_name("test.txt"), "w") as f:
                  raise ValueError()
                  f.write("Test")

    Instead of creating the file "test.txt" at the beginning and filling it with content
    later (which will never happen because of the exception thrown, which makes the file
    existing but the task actually not finished), the file will be written to a temporary
    file first and copied to its final location at the end of the run function (but only if there
    was no error).

    Warning:
        The decorator only edits the function :meth:`b2luigi.Task.get_output_file_name`. If you are using
        the output directly, you have to take care of using the temporary path correctly by yourself!

    """

    @wraps(run_function)
    def run(self, *args, **kwargs):
        with TemporaryFileContextManager(self):
            run_function(self, *args, **kwargs)

    return run
